# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Simple tests for the Network class.
"""

from functools import partial
from itertools import islice, product, repeat
from multiprocessing import get_context, cpu_count

import pytest
import numpy as np
import scipy.sparse as sp

from pyunicorn import Network


# -----------------------------------------------------------------------------
# utilities
# -----------------------------------------------------------------------------


def compare_results(desired, actual, rev_perm=None):
    assert type(desired) is type(actual)
    if isinstance(desired, dict):
        _ = [compare_results(*rs, rev_perm=rev_perm)
             for rs in zip(desired.values(), actual.values())]
    else:
        if sp.issparse(desired):
            desired, actual = desired.A, actual.A
        if isinstance(desired, np.ndarray):
            actual = actual[rev_perm]
            if len(actual.shape) == 2:
                actual = actual[:, rev_perm]
        assert np.allclose(desired, actual, atol=1e-06)


def compare_measures(orig, pnets, rev_perms, tasks):
    for (measure, i) in tasks:
        method, args = (measure, ()) if isinstance(measure, str) else measure
        compare_results(getattr(orig, method)(*args),
                        getattr(pnets[i], method)(*args), rev_perms[i])


def compare_permutations(net, permutations, measures):
    pnets, rev_perms = zip(
        *((net.permuted_copy(p), p.argsort()) for p in
          map(np.random.permutation, repeat(net.N, permutations))))
    tasks = list(product(measures, range(permutations)))
    cores = cpu_count()
    with get_context("spawn").Pool() as pool:
        pool.map(partial(compare_measures, net, pnets, rev_perms),
                 (list(islice(tasks, c, None, cores)) for c in range(cores)))
        pool.close()
        pool.join()


def compare_nsi(net, nsi_measures):
    net_copies = [net.splitted_copy(node=i) for i in range(net.N)]
    for nsi_measure in nsi_measures:
        if isinstance(nsi_measure, tuple):
            kwargs = nsi_measure[1]
            nsi_measure = nsi_measure[0]
        else:
            kwargs = {}
        print(nsi_measure)
        for i, netc in enumerate(net_copies):
            # test for invariance of old nodes
            assert np.allclose(getattr(netc, nsi_measure)(**kwargs)[0:net.N],
                               getattr(net, nsi_measure)(**kwargs))
            # test for invariance of origianl and new splitted node
            assert np.allclose(getattr(netc, nsi_measure)(**kwargs)[i],
                               getattr(netc, nsi_measure)(**kwargs)[-1])


# -----------------------------------------------------------------------------
# stability
# -----------------------------------------------------------------------------


def test_int_overflow():
    """
    Avoid integer overflow in scipy.sparse representation.
    """
    for n in [10, 200, 2000, 33000]:
        adj = sp.lil_matrix((n, n), dtype=np.int8)
        adj[0, 1:] = 1
        deg = Network(adjacency=adj).degree()
        assert (deg.min(), deg.max()) == (0, n - 1)


# -----------------------------------------------------------------------------
# consistency
# -----------------------------------------------------------------------------


def test_permutations():
    """
    Permutation invariance of topological information.
    """
    compare_permutations(
        Network.SmallTestNetwork(), 3, [
            "degree", "indegree", "outdegree", "nsi_degree", "nsi_indegree",
            "nsi_outdegree", "nsi_average_neighbors_degree",
            "nsi_max_neighbors_degree", "undirected_adjacency", "laplacian",
            "nsi_laplacian", "local_clustering", "global_clustering",
            "transitivity", ("higher_order_transitivity", [4]),
            ("local_cliquishness", [4]), ("local_cliquishness", [5]),
            "nsi_twinness", "assortativity", "nsi_local_clustering",
            "nsi_global_clustering", "nsi_transitivity",
            "nsi_local_soffer_clustering", "path_lengths",
            "average_path_length", "nsi_average_path_length", "diameter",
            "matching_index", "link_betweenness", "betweenness",
            "eigenvector_centrality", "nsi_eigenvector_centrality", "pagerank",
            "closeness", "nsi_closeness", "nsi_harmonic_closeness",
            "nsi_exponential_closeness", "arenas_betweenness",
            "nsi_arenas_betweenness", "newman_betweenness",
            "nsi_newman_betweenness", "global_efficiency",
            "nsi_global_efficiency", "distance_based_measures",
            "local_vulnerability", "coreness", "msf_synchronizability",
            "spreading", "nsi_spreading"
        ])


def test_nsi():
    """
    Consistency of nsi measures with splitted network copies
    """
    dnw = Network.SmallDirectedTestNetwork()
    nw = Network.SmallTestNetwork()

    nsi_measures = ["nsi_degree", "nsi_indegree", "nsi_outdegree",
                    "nsi_closeness", "nsi_harmonic_closeness",
                    "nsi_exponential_closeness", "nsi_arenas_betweenness",
                    "nsi_spreading",
                    "nsi_local_cyclemotif_clustering",
                    "nsi_local_midmotif_clustering",
                    "nsi_local_inmotif_clustering",
                    "nsi_local_outmotif_clustering",
                    ("nsi_degree", {"key": "link_weights"}),
                    ("nsi_indegree", {"key": "link_weights"}),
                    ("nsi_outdegree", {"key": "link_weights"}),
                    ("nsi_local_cyclemotif_clustering",
                     {"key": "link_weights"}),
                    ("nsi_local_midmotif_clustering",
                     {"key": "link_weights"}),
                    ("nsi_local_inmotif_clustering",
                     {"key": "link_weights"}),
                    ("nsi_local_outmotif_clustering",
                     {"key": "link_weights"})]

    nsi_undirected_measures = ["nsi_local_clustering",
                               "nsi_average_neighbors_degree",
                               "nsi_max_neighbors_degree",
                               "nsi_eigenvector_centrality",
                               "nsi_local_clustering",
                               "nsi_local_soffer_clustering",
                               "nsi_newman_betweenness"]

    compare_nsi(dnw, nsi_measures)
    compare_nsi(nw, nsi_measures + nsi_undirected_measures)

# -----------------------------------------------------------------------------
# Class member tests with TestNetwork
# -----------------------------------------------------------------------------


def test_init():
    Network(adjacency=[[0, 1, 0, 0, 0, 0], [1, 0, 1, 0, 0, 1],
                       [0, 1, 0, 1, 1, 0], [0, 0, 1, 0, 1, 0],
                       [0, 0, 1, 1, 0, 1], [0, 1, 0, 0, 1, 0]])
    assert True


def test_str(capsys):
    print(Network.SmallTestNetwork())
    out, err = capsys.readouterr()
    out_ref = "Network: undirected, 6 nodes, 7 links, link density 0.467.\n"
    assert out == out_ref


def test_len():
    assert np.allclose(len(Network.SmallTestNetwork()), 6)


def test_undirected_copy(capsys):
    net = Network(adjacency=[[0, 1], [0, 0]], directed=True)

    print(net)
    out1, err = capsys.readouterr()
    out1_ref = "Network: directed, 2 nodes, 1 links, link density 0.500.\n"
    assert out1 == out1_ref

    print(net.undirected_copy())
    out2, err = capsys.readouterr()
    out2_ref = "Network: undirected, 2 nodes, 1 links, link density 1.000.\n"
    assert out2 == out2_ref


def test_splitted_copy(capsys):
    net = Network.SmallTestNetwork()
    net2 = net.splitted_copy(node=5, proportion=0.2)
    print(net2)
    out, err = capsys.readouterr()
    out_ref = "Network: undirected, 7 nodes, 9 links, link density 0.429.\n"
    assert out == out_ref

    nw1_ref = [1.5, 1.7, 1.9, 2.1, 2.3, 2.5]
    nw2_ref = [1.5, 1.7, 1.9, 2.1, 2.3, 2., 0.5]

    assert np.allclose(net.node_weights, nw1_ref)
    assert np.allclose(net2.node_weights, nw2_ref)


def test_adjacency():
    adj_ref = np.array([[0, 0, 0, 1, 1, 1], [0, 0, 1, 1, 1, 0],
                        [0, 1, 0, 0, 1, 0], [1, 1, 0, 0, 0, 0],
                        [1, 1, 1, 0, 0, 0], [1, 0, 0, 0, 0, 0]])
    assert np.array_equal(Network.SmallTestNetwork().adjacency, adj_ref)


def test_set_adjacency(capsys):
    net = Network.SmallTestNetwork()
    net.adjacency = [[0, 1], [1, 0]]
    print(net)
    out, err = capsys.readouterr()
    out_ref = "Network: undirected, 2 nodes, 1 links, link density 1.000.\n"
    assert out == out_ref


def test_set_node_weights(capsys):
    net = Network.SmallTestNetwork()
    nw_ref = [1.5, 1.7, 1.9, 2.1, 2.3, 2.5]
    assert np.allclose(net.node_weights, nw_ref)

    net.node_weights = [1, 1, 1, 1, 1, 1]
    nw_ref = [1., 1., 1., 1., 1., 1.]
    assert np.allclose(net.node_weights, nw_ref)


def test_ErdosRenyi(capsys):
    print(Network.Model("ErdosRenyi", n_nodes=10, n_links=18))
    out, err = capsys.readouterr()
    out_ref = "Generating Erdos-Renyi random graph with 10 " + \
              "nodes and 18 links...\n" + \
              "Network: undirected, 10 nodes, 18 links, link density 0.400.\n"
    assert out == out_ref


def test_BarabasiAlbert_igraph():
    net = Network.Model("BarabasiAlbert_igraph", n_nodes=100, n_links_each=1)
    assert np.allclose(net.link_density, 0.02)


def test_ConfigurationModel():
    net = Network.Model("Configuration", degrees=[3 for _ in range(0, 1000)])
    assert int(round(net.degree().mean())) == 3


def test_WattsStrogatz():
    net = Network.Model("WattsStrogatz", N=100, k=2, p=0.1)
    assert int(round(net.degree().mean())) == 4


def test_randomly_rewire(capsys):
    net = Network.SmallTestNetwork()
    net.randomly_rewire(iterations=10)
    out, err = capsys.readouterr()
    out_ref = "Randomly rewiring the network,preserving " + \
              "the degree sequence...\n"
    assert out == out_ref
    print(net)
    out, err = capsys.readouterr()
    out_ref = "Network: undirected, 6 nodes, 7 links, link density 0.467.\n"
    assert out == out_ref


def test_edge_list():
    edges = Network.SmallTestNetwork().edge_list()[:8]
    edges_ref = [[0, 3], [0, 4], [0, 5], [1, 2],
                 [1, 3], [1, 4], [2, 1], [2, 4]]
    assert np.array_equal(edges, edges_ref)


def test_undirected_adjacency():
    net = Network(adjacency=[[0, 1], [0, 0]], directed=True)
    adj_ref = [[0, 1], [1, 0]]
    assert np.array_equal(net.undirected_adjacency().A, adj_ref)


def test_laplacian():
    lap_ref = np.array([[3, 0, 0, -1, -1, -1], [0, 3, -1, -1, -1, 0],
                        [0, -1, 2, 0, -1, 0], [-1, -1, 0, 2, 0, 0],
                        [-1, -1, -1, 0, 3, 0], [-1, 0, 0, 0, 0, 1]])
    assert np.allclose(Network.SmallTestNetwork().laplacian(), lap_ref)


def test_nsi_laplacian():
    nsi_lap_ref = np.array([[6.9, 0., 0., -2.1, -2.3, -2.5],
                            [0., 6.3, -1.9, -2.1, -2.3, 0.],
                            [0., -1.7, 4., 0., -2.3, 0.],
                            [-1.5, -1.7, 0., 3.2, 0., 0.],
                            [-1.5, -1.7, -1.9, 0., 5.1, 0.],
                            [-1.5, 0., 0., 0., 0., 1.5]])
    assert np.allclose(Network.SmallTestNetwork().nsi_laplacian(), nsi_lap_ref)


def test_degree():
    deg = Network.SmallTestNetwork().degree()
    deg_ref = np.array([3, 3, 2, 2, 3, 1])
    assert (deg == deg_ref).all()


def test_indegree():
    deg = Network.SmallDirectedTestNetwork().indegree()
    deg_ref = np.array([2, 2, 2, 1, 1, 0])
    assert (deg == deg_ref).all()


def test_outdegree():
    deg = Network.SmallDirectedTestNetwork().outdegree()
    deg_ref = np.array([2, 2, 0, 1, 2, 1])
    assert (deg == deg_ref).all()


def test_bildegree():
    deg = Network.SmallDirectedTestNetwork().bildegree()
    deg_ref = np.array([0, 0, 0, 0, 0, 0], dtype=np.int16)
    assert (deg == deg_ref).all()

    net = Network.SmallTestNetwork()
    assert (net.bildegree() == net.degree()).all()


def test_nsi_degree():
    net = Network.SmallTestNetwork()

    deg_ref = np.array([8.4, 8., 5.9, 5.3, 7.4, 4.])
    assert np.allclose(net.nsi_degree(), deg_ref)

    deg_ref = np.array([8.4, 8., 5.9, 5.3, 7.4, 4., 4.])
    assert np.allclose(net.splitted_copy().nsi_degree(), deg_ref)

    deg_ref = np.array([3.2, 3., 1.95, 1.65, 2.7, 1.])
    assert np.allclose(net.nsi_degree(typical_weight=2.0), deg_ref)

    deg_ref = np.array([3.2, 3., 1.95, 1.65, 2.7, 1., 1.])
    assert np.allclose(net.splitted_copy().nsi_degree(typical_weight=2.0),
                       deg_ref)


@pytest.mark.parametrize("tw, exp, exp_split", [
    (None,
     np.array([6.3, 5.3, 5.9, 3.6, 4., 2.5]),
     np.array([6.3, 5.3, 5.9, 3.6, 4., 2.5, 2.5])),
    (2.,
     np.array([2.15, 1.65, 1.95, 0.8, 1., 0.25]),
     np.array([2.15, 1.65, 1.95, 0.8, 1., 0.25, 0.25]))
    ])
def test_nsi_indegree(tw, exp, exp_split):
    net = Network.SmallDirectedTestNetwork()
    assert np.allclose(net.nsi_indegree(typical_weight=tw), exp)
    assert np.allclose(
        net.splitted_copy().nsi_indegree(typical_weight=tw), exp_split)


@pytest.mark.parametrize("tw, exp, exp_split", [
    (None,
     np.array([5.3, 5.9, 1.9, 3.8, 5.7, 4.]),
     np.array([5.3, 5.9, 1.9, 3.8, 5.7, 4., 4.])),
    (2.,
     np.array([1.65, 1.95, -0.05, 0.9, 1.85, 1.]),
     np.array([1.65, 1.95, -0.05, 0.9, 1.85, 1., 1.]))
    ])
def test_nsi_outdegree(tw, exp, exp_split):
    net = Network.SmallDirectedTestNetwork()
    assert np.allclose(net.nsi_outdegree(typical_weight=tw), exp)
    assert np.allclose(net.splitted_copy().nsi_outdegree(typical_weight=tw),
                       exp_split)


@pytest.mark.parametrize("tw, exp, exp_split", [
    (None,
     np.array([1.5, 1.7, 1.9, 2.1, 2.3, 2.5]),
     np.array([1.5, 1.7, 1.9, 2.1, 2.3, 2.5, 2.5])),
    (2.,
     np.array([-0.25, -0.15, -0.05,  0.05,  0.15,  0.25]),
     np.array([-0.25, -0.15, -0.05,  0.05,  0.15,  0.25,  0.25]))
    ])
def test_nsi_bildegree(tw, exp, exp_split):
    net = Network.SmallDirectedTestNetwork()
    assert np.allclose(net.nsi_bildegree(typical_weight=tw), exp)
    assert np.allclose(net.splitted_copy().nsi_bildegree(typical_weight=tw),
                       exp_split)


def test_degree_distribution():
    dist = Network.SmallTestNetwork().degree_distribution()
    dist_ref = np.array([0.16666667, 0.33333333, 0.5])
    assert np.allclose(dist, dist_ref)


def test_indegree_distribution():
    dist = Network.SmallTestNetwork().indegree_distribution()
    dist_ref = np.array([0.16666667, 0.33333333, 0.5])
    assert np.allclose(dist, dist_ref)


def test_outdegree_distribution():
    dist = Network.SmallTestNetwork().outdegree_distribution()
    dist_ref = np.array([0.16666667, 0., 0.33333333, 0.5])
    assert np.allclose(dist, dist_ref)


def test_degree_cdf():
    cdf_ref = np.array([1., 0.83333333, 0.5])
    assert np.allclose(Network.SmallTestNetwork().degree_cdf(), cdf_ref)


def test_indegree_cdf():
    cdf_ref = np.array([1., 0.83333333, 0.83333333, 0.5])
    assert np.allclose(Network.SmallTestNetwork().indegree_cdf(), cdf_ref)


def test_outdegree_cdf():
    cdf_ref = np.array([1., 0.83333333, 0.83333333, 0.5])
    assert np.allclose(Network.SmallTestNetwork().outdegree_cdf(), cdf_ref)


def test_nsi_degree_histogram():
    hist = Network.SmallTestNetwork().nsi_degree_histogram()
    hist_ref = (np.array([0.33333333, 0.16666667, 0.5]),
                np.array([0.11785113, 0.16666667, 0.09622504]),
                np.array([4., 5.46666667, 6.93333333]))
    assert np.allclose(hist, hist_ref)


def test_nsi_degree_cumulative_histogram():
    res = Network.SmallTestNetwork().nsi_degree_cumulative_histogram()
    exp = (np.array([1., 0.66666667, 0.5]),
           np.array([4., 5.46666667, 6.93333333]))
    assert np.allclose(res, exp)


def test_average_neighbors_degree():
    res = Network.SmallTestNetwork().average_neighbors_degree()
    exp = np.array([2., 2.33333333, 3., 3., 2.66666667, 3.])
    assert np.allclose(res, exp)


def test_max_neighbors_degree():
    res = Network.SmallTestNetwork().max_neighbors_degree()
    exp = np.array([3, 3, 3, 3, 3, 3])
    assert (res == exp).all()


def test_nsi_average_neighbors_degree():
    net = Network.SmallTestNetwork()

    res = net.nsi_average_neighbors_degree()
    exp = np.array([6.0417, 6.62, 7.0898, 7.0434, 7.3554, 5.65])
    assert np.allclose(res, exp)

    res = net.splitted_copy().nsi_average_neighbors_degree()
    exp = np.array([6.0417, 6.62, 7.0898, 7.0434, 7.3554, 5.65, 5.65])
    assert np.allclose(res, exp)


def test_nsi_max_neighbors_degree():
    res = Network.SmallTestNetwork().nsi_max_neighbors_degree()
    exp = np.array([8.4, 8., 8., 8.4, 8.4, 8.4])
    assert np.allclose(res, exp)


def test_local_clustering():
    res = Network.SmallTestNetwork().local_clustering()
    exp = np.array([0., 0.33333333, 1., 0., 0.33333333, 0.])
    assert np.allclose(res, exp)


def test_global_clustering():
    res = Network.SmallTestNetwork().global_clustering()
    exp = 0.27777777
    assert np.allclose(res, exp)


def test_local_cyclemotif_clustering():
    res = Network.SmallDirectedTestNetwork().local_cyclemotif_clustering()
    exp = np.array([0.25, 0.25, 0., 0., 0.5, 0.])
    assert np.allclose(res, exp)


def test_local_midmotif_clustering():
    res = Network.SmallDirectedTestNetwork().local_midmotif_clustering()
    exp = np.array([0., 0., 0., 1., 0.5, 0.])
    assert np.allclose(res, exp)


def test_local_inmotif_clustering():
    res = Network.SmallDirectedTestNetwork().local_inmotif_clustering()
    exp = np.array([0., 0.5, 0.5, 0., 0., 0.])
    assert np.allclose(res, exp)


def test_local_outmotif_clustering():
    res = Network.SmallDirectedTestNetwork().local_outmotif_clustering()
    exp = np.array([0.5, 0.5, 0., 0., 0., 0.])
    assert np.allclose(res, exp)


@pytest.mark.parametrize("tw, exp, exp_split", [
    (None,
     np.array([0.18448637, 0.20275024, 0.3220339,
               0.32236842, 0.34385965, 0.625]),
     np.array([0.18448637, 0.20275024, 0.3220339,
               0.32236842, 0.34385965, 0.625, 0.20275024])),
    (2.,
     np.array([0.3309814,  0.29011913,  0.05236908,
               -0.0260989,  0.22417582, -0.13636364]),
     np.array([0.3309814,  0.29011913,  0.05236908,
               -0.0260989,  0.22417582, -0.13636364,  0.29011913]))
    ])
def test_nsi_local_cyclemotif_clustering(tw, exp, exp_split):
    net = Network.SmallDirectedTestNetwork()
    assert np.allclose(net.nsi_local_cyclemotif_clustering(typical_weight=tw),
                       exp)
    assert np.allclose(
        net.splitted_copy(node=1).nsi_local_cyclemotif_clustering(
            typical_weight=tw),
        exp_split)


def test_nsi_local_midmotif_clustering():
    net = Network.SmallDirectedTestNetwork()

    res = net.nsi_local_midmotif_clustering()
    exp = np.array([0.45372866, 0.51646946, 1., 1., 0.88815789, 1.])
    assert np.allclose(res, exp)

    res = net.splitted_copy(node=4).local_midmotif_clustering()
    exp = np.array([0., 0., 0., 1., 0.8, 0., 0.8])
    assert np.allclose(res, exp)


def test_nsi_local_inmotif_clustering():
    net = Network.SmallDirectedTestNetwork()

    res = net.nsi_local_inmotif_clustering()
    exp = np.array([0.52884858, 0.66998932, 0.66934789, 0.75694444, 0.755625,
                    1.])
    assert np.allclose(res, exp)

    res = net.splitted_copy(node=1).nsi_local_inmotif_clustering()
    exp = np.array([0.52884858, 0.66998932, 0.66934789, 0.75694444, 0.755625,
                    1., 0.66998932])
    assert np.allclose(res, exp)


def test_nsi_local_outmotif_clustering():
    net = Network.SmallDirectedTestNetwork()

    res = net.nsi_local_outmotif_clustering()
    exp = np.array([0.66998932, 0.66934789, 1., 0.75277008, 0.58387196,
                    0.765625])
    assert np.allclose(res, exp)

    res = net.splitted_copy(node=1).nsi_local_outmotif_clustering()
    exp = np.array([0.66998932, 0.66934789, 1., 0.75277008, 0.58387196,
                    0.765625, 0.66934789])
    assert np.allclose(res, exp)


def test_transitivity():
    res = Network.SmallTestNetwork().transitivity()
    exp = 0.27272727
    assert np.allclose(res, exp)


def test_weighted_local_clustering():
    res = Network.weighted_local_clustering(
        weighted_A=[[0., 0., 0., 0.55, 0.65, 0.75],
                    [0., 0., 0.63, 0.77, 0.91, 0.],
                    [0., 0.63, 0., 0., 1.17, 0.],
                    [0.55, 0.77, 0., 0., 0., 0.],
                    [0.65, 0.91, 1.17, 0., 0., 0.],
                    [0.75, 0., 0., 0., 0., 0.]])
    exp = np.array([0., 0.21487603, 0.35388889, 0., 0.15384615, 0.])
    assert np.allclose(res, exp)


def test_nsi_twinness():
    net = Network.SmallDirectedTestNetwork()

    res = net.nsi_twinness()
    exp = np.array([[0.12931034, 0.45689655, 0., 0.31034483, 0., 0.],
                    [0., 0.15178571, 0.52678571, 0., 0.35714286, 0.],
                    [0., 0., 0.24358974, 0., 0., 0.],
                    [0., 0.33928571, 0., 0.28378378, 0., 0.],
                    [0.32758621, 0., 0.43298969, 0., 0.2371134, 0.],
                    [0.34482759, 0., 0., 0., 0., 0.38461538]])
    assert np.allclose(res, exp)

    res = net.splitted_copy().nsi_twinness()
    exp = np.array([[0.12931034, 0.45689655, 0., 0.31034483, 0., 0., 0.],
                    [0., 0.15178571, 0.52678571, 0., 0.35714286, 0., 0.],
                    [0., 0., 0.24358974, 0., 0., 0., 0.],
                    [0., 0.33928571, 0., 0.28378378, 0., 0., 0.],
                    [0.32758621, 0., 0.43298969, 0., 0.2371134, 0., 0.],
                    [0.34482759, 0., 0., 0., 0., 0.38461538, 0.38461538],
                    [0.34482759, 0., 0., 0., 0., 0.38461538, 0.38461538]])
    assert np.allclose(res, exp)


def test_assortativity():
    res = Network.SmallTestNetwork().assortativity()
    exp = -0.47368421
    assert np.allclose(res, exp)


@pytest.mark.parametrize("tw, exp, exp_split", [
    (None,
     np.array([0.55130385, 0.724375, 1., 0.81844073, 0.80277575, 1.]),
     np.array([0.55130385, 0.724375, 1., 0.81844073, 0.80277575, 1., 1.])),
    (3.,
     np.array([-1.44290123, -0.764, 1., 4.16770186, -0.75324675, 1.]),
     np.array([-1.44290123, -0.764, 1., 4.16770186, -0.75324675, 1., 1.]))
    ])
def test_nsi_local_clustering(tw, exp, exp_split):
    net = Network.SmallTestNetwork()

    assert np.allclose(net.nsi_local_clustering(typical_weight=tw), exp)
    assert np.allclose(
        net.splitted_copy().nsi_local_clustering(typical_weight=tw),
        exp_split)


def test_nsi_global_clustering():
    res = Network.SmallTestNetwork().nsi_global_clustering()
    exp = 0.83529192
    assert np.allclose(res, exp)


def test_nsi_local_soffer_clustering():
    net = Network.SmallTestNetwork()

    res = net.nsi_local_soffer_clustering()
    exp = np.array([0.76650246, 0.87537764, 1., 0.81844073, 0.84685032, 1.])
    assert np.allclose(res, exp)

    res = net.splitted_copy().nsi_local_soffer_clustering()
    exp = np.array([0.76650246, 0.87537764, 1., 0.81844073, 0.84685032, 1.,
                    1.])
    assert np.allclose(res, exp)


def test_path_lengths():
    res = Network.SmallTestNetwork().path_lengths()
    exp = np.array([[0., 2., 2., 1., 1., 1.],
                    [2., 0., 1., 1., 1., 3.],
                    [2., 1., 0., 2., 1., 3.],
                    [1., 1., 2., 0., 2., 2.],
                    [1., 1., 1., 2., 0., 2.],
                    [1., 3., 3., 2., 2., 0.]])
    assert np.allclose(res, exp)


def test_average_path_length():
    res = Network.SmallTestNetwork().average_path_length()
    exp = 1.66666667
    assert np.allclose(res, exp)


def test_nsi_average_path_length():
    net = Network.SmallTestNetwork()

    res = net.nsi_average_path_length()
    exp = 1.60027778
    assert np.allclose(res, exp)

    res = net.splitted_copy().nsi_average_path_length()
    exp = 1.60027778
    assert np.allclose(res, exp)


def test_diameter():
    res = Network.SmallTestNetwork().diameter()
    exp = 3
    assert np.allclose(res, exp)


def test_matching_index():
    res = Network.SmallTestNetwork().matching_index()
    exp = np.array([[1., 0.5, 0.25, 0., 0., 0.],
                    [0.5, 1., 0.25, 0., 0.2, 0.],
                    [0.25, 0.25, 1., 0.33333333, 0.25, 0.],
                    [0., 0., 0.33333333, 1., 0.66666667, 0.5],
                    [0., 0.2, 0.25, 0.66666667, 1., 0.33333333],
                    [0., 0., 0., 0.5, 0.33333333, 1.]])
    assert np.allclose(res, exp)


def test_link_betweenness():
    res = Network.SmallTestNetwork().link_betweenness()
    exp = np.array([[0., 0., 0., 3.5, 5.5, 5.],
                    [0., 0., 2., 3.5, 2.5, 0.],
                    [0., 2., 0., 0., 3., 0.],
                    [3.5, 3.5, 0., 0., 0., 0.],
                    [5.5, 2.5, 3., 0., 0., 0.],
                    [5., 0., 0., 0., 0., 0.]])
    assert np.allclose(res, exp)


def test_edge_betweenness():
    res = Network.SmallTestNetwork().edge_betweenness()
    exp = np.array([[0., 0., 0., 3.5, 5.5, 5.],
                    [0., 0., 2., 3.5, 2.5, 0.],
                    [0., 2., 0., 0., 3., 0.],
                    [3.5, 3.5, 0., 0., 0., 0.],
                    [5.5, 2.5, 3., 0., 0., 0.],
                    [5., 0., 0., 0., 0., 0.]])
    assert np.allclose(res, exp)


def test_betweenness():
    res = Network.SmallTestNetwork().betweenness()
    exp = np.array([4.5, 1.5, 0., 1., 3., 0.])
    assert np.allclose(res, exp)


def test_interregional_betweenness():
    net = Network.SmallTestNetwork()
    res = net.interregional_betweenness(sources=[2], targets=[3, 5])
    exp = np.array([1., 1., 0., 0., 1., 0.])
    assert np.allclose(res, exp)

    res = net.interregional_betweenness(sources=range(0, 6),
                                        targets=range(0, 6))
    exp = np.array([9., 3., 0., 2., 6., 0.])
    assert np.allclose(res, exp)


def test_nsi_interregional_betweenness():
    res = Network.SmallTestNetwork().nsi_interregional_betweenness(
        sources=[2], targets=[3, 5])
    exp = np.array([3.16666689, 2.34705893, 0., 0., 2.06521743, 0.])
    assert np.allclose(res, exp)


@pytest.mark.parametrize("parallelize", [False, True])
def test_nsi_betweenness(parallelize):
    net = Network.SmallTestNetwork()

    res = net.nsi_betweenness(parallelize=parallelize)
    exp = np.array([29.68541738, 7.7128677, 0., 3.09090906, 9.69960462, 0.])
    assert np.allclose(res, exp)

    res = net.splitted_copy().nsi_betweenness(parallelize=parallelize)
    exp = np.append(exp, [0.])
    assert np.allclose(res, exp)


def test_eigenvector_centrality():
    res = Network.SmallTestNetwork().eigenvector_centrality()
    exp = np.array([0.7895106, 0.97303126, 0.77694188, 0.69405519, 1.,
                    0.31089413])
    assert np.allclose(res, exp)


def test_nsi_eigenvector_centrality():
    net = Network.SmallTestNetwork()

    res = net.nsi_eigenvector_centrality()
    exp = np.array([0.80454492, 1., 0.80931481, 0.61787145, 0.98666885,
                    0.28035747])
    assert np.allclose(res, exp)

    res = net.splitted_copy().nsi_eigenvector_centrality()
    exp = np.array([0.80454492, 1., 0.80931481, 0.61787145, 0.98666885,
                    0.28035747, 0.28035747])
    assert np.allclose(res, exp)


def test_pagerank():
    res = Network.SmallTestNetwork().pagerank()
    exp = np.array([0.21836231, 0.20440819, 0.14090543, 0.14478497, 0.20466978,
                    0.08686932])
    assert np.allclose(res, exp)


def test_closeness():
    res = Network.SmallTestNetwork().closeness()
    exp = np.array([0.71428571, 0.625, 0.55555556, 0.625, 0.71428571,
                    0.45454545])
    assert np.allclose(res, exp)


def test_nsi_closeness():
    net = Network.SmallTestNetwork()

    res = net.nsi_closeness()
    exp = np.array([0.76923077, 0.64864865, 0.58252427, 0.64171123, 0.72289157,
                    0.50847458])
    assert np.allclose(res, exp)

    res = net.splitted_copy().nsi_closeness()
    exp = np.array([0.76923077, 0.64864865, 0.58252427, 0.64171123, 0.72289157,
                    0.50847458, 0.50847458])
    assert np.allclose(res, exp)


def test_nsi_harmonic_closeness():
    net = Network.SmallTestNetwork()

    res = net.nsi_harmonic_closeness()
    exp = np.array([0.85, 0.79861111, 0.71111111, 0.72083333, 0.80833333,
                    0.61666667])
    assert np.allclose(res, exp)

    res = net.splitted_copy().nsi_harmonic_closeness()
    exp = np.array([0.85, 0.79861111, 0.71111111, 0.72083333, 0.80833333,
                    0.61666667, 0.61666667])
    assert np.allclose(res, exp)


def test_nsi_exponential_closeness():
    net = Network.SmallTestNetwork()

    res = net.nsi_exponential_closeness()
    exp = np.array([0.425, 0.390625, 0.346875, 0.36041667, 0.40416667,
                    0.29583333])
    assert np.allclose(res, exp)

    res = net.splitted_copy().nsi_exponential_closeness()
    exp = np.array([0.425, 0.390625, 0.346875, 0.36041667, 0.40416667,
                    0.29583333, 0.29583333])
    assert np.allclose(res, exp)


def test_arenas_betweenness():
    res = Network.SmallTestNetwork().arenas_betweenness()
    exp = np.array([50.18181818, 50.18181818, 33.45454545, 33.45454545,
                    50.18181818, 16.72727273])
    assert np.allclose(res, exp)


def test_nsi_arenas_betweenness():
    net = Network.SmallTestNetwork()

    res = net.nsi_arenas_betweenness()
    exp = np.array([20.58135241, 29.21033898, 27.00747741, 19.5433536,
                    25.28490117, 24.84826305])
    assert np.allclose(res, exp)

    res = net.splitted_copy().nsi_arenas_betweenness()
    exp = np.array([20.58135241, 29.21033898, 27.00747741, 19.5433536,
                    25.28490117, 24.84826305, 24.84826305])
    assert np.allclose(res, exp)

    res = net.nsi_arenas_betweenness(exclude_neighbors=False)
    exp = np.array([44.53512333, 37.40578733, 27.00747741, 21.77360559,
                    31.32557606, 24.84826305])
    assert np.allclose(res, exp)

    res = net.nsi_arenas_betweenness(stopping_mode="twinness")
    exp = np.array([22.61533156, 41.23139296, 38.64105931, 28.61953314,
                    38.58242175, 30.29941829])
    assert np.allclose(res, exp)


def test_newman_betweenness():
    res = Network.SmallTestNetwork().newman_betweenness()
    exp = np.array([4.1818182, 3.41818185, 2.5090909, 3.0181818, 3.60000002,
                    2.])
    assert np.allclose(res, exp)


def test_nsi_newman_betweenness():
    net = Network.SmallTestNetwork()

    res = net.nsi_newman_betweenness()
    exp = np.array([0.40476082, 0., 0.85212808, 3.33573728, 1.36618345, 0.])
    assert np.allclose(res, exp)

    res = net.splitted_copy().nsi_newman_betweenness()
    exp = np.array([0.40476082, 0., 0.85212808, 3.33573728, 1.36618345,
                    0., 0.])
    assert np.allclose(res, exp)

    res = net.nsi_newman_betweenness(add_local_ends=True)
    exp = np.array([131.44476082, 128., 107.64212808, 102.44573728,
                    124.20618345, 80.])
    assert np.allclose(res, exp)

    res = net.splitted_copy().nsi_newman_betweenness(add_local_ends=True)
    exp = np.array([131.44476082, 128., 107.64212808, 102.44573728,
                    124.20618345, 80., 80.])
    assert np.allclose(res, exp)


def test_global_efficiency():
    res = Network.SmallTestNetwork().global_efficiency()
    exp = 0.71111111
    assert np.allclose(res, exp)


def test_nsi_global_efficiency():
    res = Network.SmallTestNetwork().nsi_global_efficiency()
    exp = 0.74152777
    assert np.allclose(res, exp)


def test_local_vulnerability():
    res = Network.SmallTestNetwork().local_vulnerability()
    exp = np.array([0.296875, 0.0625, -0.03125, -0.0078125, 0.09765625,
                    -0.125])
    assert np.allclose(res, exp)


def test_coreness():
    res = Network.SmallTestNetwork().coreness()
    exp = np.array([2, 2, 2, 2, 2, 1])
    assert (res == exp).all()


def test_msf_synchronizability():
    res = Network.SmallTestNetwork().msf_synchronizability()
    exp = 6.77842586
    assert np.allclose(res, exp)



# Copyright (C) 2014 SWIPO Project
#
# Authors (this file):
#   Stefan Schinkel <stefan.schinkel@gmail.com>

"""
Utils needed for Unit Tests in resistive networks
"""
import numpy as np


def makeNW(idI, idJ, val=None):
    """Construct NW from edges
    """

    # idI and IdJ must be lists!
    idI = list(idI)
    idJ = list(idJ)

    # get N
    N = max(list(idI)+list(idJ))

    # empty matrix
    nw = np.zeros((N+1, N+1))

    # fill values or ones
    if val is None:
        val = [1] * len(idI)
    elif len(val) == 1:
        val = val*len(idI)

    for i, j, v in zip(idI, idJ, val):
        nw[i, j] = v
        nw[j, i] = v

    return nw


def parallelCopy(nw, a, b):
    """ copy a graph of N nodes with a given adjacency
    matrix and attach the copy linking node them at node
    a and b (parallel circuit)
    """

    if sum(np.diag(nw)) > 0:
        print("Graph has selfloops")
        print("doesn't work yet")
        raise NotImplementedError

        # at this point, have to copy the diagonal
        # and fill that to the copy
        # np.fill_diagonal(nw,0)

    # get size of NW
    N = len(nw)

    # get the current adjacency and values
    # of the upper triangle only (has to be symmetric)
    i0, j0 = np.nonzero(np.triu(nw))
    val = []
    for ii, jj in zip(i0, j0):
        val.append(nw[ii, jj])

    # increment i indeces and remap
    ti = i0+N
    ti[ti == a+N] = a
    ii = np.append(i0, ti)

    # same for j
    tj = j0+N
    tj[tj == b+N] = b
    jj = np.append(j0, tj)

    # compute return value and adjust
    # the values for direct links
    nw2 = makeNW(ii, jj, val*2)
    if nw[a, b] != 0:
        nw2[a, b] = nw[a, b]/2
        nw2[b, a] = nw[a, b]/2

    return nw2


def serialCopy(nw):
    """ Copy a graph and attach a copy to
    the end (serial circuit)
    """

    # get size of NW
    N = len(nw)

    # get the current adjacency and values
    # of the upper triangle only (has to be symmetric)
    i0, j0 = np.nonzero(np.triu(nw))
    val = []
    for ii, jj in zip(i0, j0):
        val.append(nw[ii, jj])

    # increment indeces by N and append
    ti = i0+N-1
    ii = np.append(i0, ti)

    # same for j
    tj = j0+N-1
    jj = np.append(j0, tj)

    return makeNW(ii, jj, val*2)


def nx2nw(G):
    """Convert networkx Graph to a plain matrix
    (ndarray)
    """
    i = []
    j = []
    for edge in G.edges():
        i.append(edge[0])
        j.append(edge[1])

    return makeNW(i, j)



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

import numpy as np
from numpy cimport ndarray

from ...core._ext.types import LAG, FIELD, DFIELD, NODE
from ...core._ext.types cimport LAG_t, FIELD_t, DFIELD_t, NODE_t

# coupling_analysis ===========================================================


def _symmetrize_by_absmax(
        ndarray[FIELD_t, ndim=2, mode='c'] similarity_matrix not None,
        ndarray[LAG_t, ndim=2, mode='c'] lag_matrix not None, int N):

    cdef:
        int i, j, I, J

    # loop over all node pairs
    for i in range(N):
        for j in range(i+1, N):
            # calculate max and argmax by comparing to
            # previous value and storing max
            if abs(similarity_matrix[i, j]) > abs(similarity_matrix[j, i]):
                I, J = i, j
            else:
                I, J = j, i
            similarity_matrix[J, I] = similarity_matrix[I, J]
            lag_matrix[J, I] = -lag_matrix[I, J]

    return similarity_matrix, lag_matrix


def _cross_correlation_max(
        ndarray[FIELD_t, ndim=3, mode='c'] array not None,
        int N, int tau_max, int corr_range):

    cdef:
        ndarray[FIELD_t, ndim=2, mode='c'] similarity_matrix = np.ones(
            (N, N), dtype=FIELD)
        ndarray[LAG_t, ndim=2, mode='c'] lag_matrix = np.zeros(
            (N, N), dtype=LAG)
        double crossij, max
        int i, j, tau, k, argmax

    # loop over all node pairs, NOT symmetric due to time shifts!
    for i in range(N):
        for j in range(N):
            if i != j:
                max = 0.0
                argmax = 0
                # loop over taus INCLUDING the last tau value
                for tau in range(tau_max + 1):
                    crossij = 0
                    # here the actual cross correlation is calculated
                    # assuming standardized arrays
                    for k in range(corr_range):
                        crossij += array[tau, i, k] * array[tau_max, j, k]
                    # calculate max and argmax by comparing to
                    # previous value and storing max
                    if abs(crossij) > abs(max):
                        max = crossij
                        argmax = tau
                similarity_matrix[i, j] = <FIELD_t> (max / corr_range)
                lag_matrix[i, j] = <LAG_t> (tau_max - argmax)

    return similarity_matrix, lag_matrix


def _cross_correlation_all(
        ndarray[FIELD_t, ndim=3, mode='c'] array not None,
        int N, int tau_max, int corr_range):

    cdef:
        int i, j, tau, k
        double crossij
        ndarray[FIELD_t, ndim=3, mode='c'] lagfuncs = np.zeros(
            (N, N, tau_max+1), dtype=FIELD)

    # loop over all node pairs, NOT symmetric due to time shifts!
    for i in range(N):
        for j in range(N):
            # loop over taus INCLUDING the last tau value
            for tau in range(tau_max + 1):
                crossij = 0
                # here the actual cross correlation is calculated
                # assuming standardized arrays
                for k in range(corr_range):
                    crossij += array[tau, i, k] * array[tau_max, j, k]
                lagfuncs[i, j, tau_max-tau] = <FIELD_t> (crossij / corr_range)

    return lagfuncs


def _get_nearest_neighbors(
        ndarray[FIELD_t, ndim=2, mode='c'] array not None,
        int dim, int T, int dim_x, int dim_y, int k):

    # Initialize
    cdef:
        int i, j, index, t, m, n, d, kxz, kyz, kz
        double dz, dxyz, dx, dy, eps, epsmax
        ndarray[NODE_t, ndim=1, mode='c'] indexfound = np.zeros(T, dtype=NODE)
        ndarray[DFIELD_t, ndim=2, mode='c'] dist = np.zeros((dim, T), dtype=DFIELD)
        ndarray[DFIELD_t, ndim=1, mode='c'] dxyzarray = np.zeros(k+1, dtype=DFIELD)
        ndarray[NODE_t, ndim=1, mode='c'] k_xz = np.zeros(T, dtype=NODE)
        ndarray[NODE_t, ndim=1, mode='c'] k_yz = np.zeros(T, dtype=NODE)
        ndarray[NODE_t, ndim=1, mode='c'] k_z = np.zeros(T, dtype=NODE)

    # Loop over time
    for i in range(T):
        # Growing cube algorithm:
        # Test if n = #(points in epsilon-environment of reference point i) > k
        # Start with epsilon for which 95% of points are inside the cube
        # for a multivariate Gaussian
        # eps increased by 2 later, also the initial eps
        eps = (k/T)**(1./dim)

        # n counts the number of neighbors
        n = 0
        while n <= k:
            # Increase cube size
            eps *= 2.
            # Start with zero again
            n = 0
            # Loop through all points
            for t in range(T):
                d = 0
                while d < dim and abs(array[d, i] - array[d, t]) < eps:
                    d += 1

                # If all distances are within eps, the point t lies
                # within eps and n is incremented
                if d == dim:
                    indexfound[n] = t
                    n += 1

        # Calculate distance to points only within epsilon environment
        # according to maximum metric
        for j in range(n):
            index = indexfound[j]

            # calculate maximum metric distance to point
            dxyz = 0.
            for d in range(dim):
                dist[d, j] = abs(array[d, i] - array[d, index])
                dxyz = max(dist[d, j], dxyz)

            # insertion-sort current distance into 'dxyzarray'
            # if it is among the currently smallest k+1 distances
            if j == 0:
                dxyzarray[j] = dxyz
            else:
                m = min(k, <int> (j-1))
                # go through previously sorted smallest distances and
                # if it is smaller than any, find slot for current distance
                while m >= 0 and dxyz < dxyzarray[m]:
                    # if it's not in the last slot already,
                    # move previously found distance to the right
                    if not m == k:
                        dxyzarray[m+1] = dxyzarray[m]
                    m -= 1

                # sort in, if a slot was found
                if not m == k:
                    dxyzarray[m+1] = dxyz

        # Epsilon of k-th nearest neighbor in joint space
        epsmax = dxyzarray[k]

        # Count neighbors within epsmax in subspaces, since the reference
        # point is included, all neighbors are at least 1
        kz = 0
        kxz = 0
        kyz = 0
        for j in range(T):

            # X-subspace
            dx = abs(array[0, i] - array[0, j])
            for d in range(1, dim_x):
                dist[d, j] = abs(array[d, i] - array[d, j])
                dx = max(dist[d, j], dx)

            # Y-subspace
            dy = abs(array[dim_x, i] - array[dim_x, j])
            for d in range(dim_x, dim_x + dim_y):
                dist[d, j] = abs(array[d, i] - array[d, j])
                dy = max(dist[d, j], dy)

            # Z-subspace, if empty, dz stays 0
            dz = 0.
            for d in range(dim_x + dim_y, dim):
                dist[d, j] = abs(array[d, i] - array[d, j])
                dz = max(dist[d, j], dz)

            # For no conditions, kz is counted up to T
            if dz < epsmax:
                kz += 1
                if dx < epsmax:
                    kxz += 1
                if dy < epsmax:
                    kyz += 1

        # Write to numpy arrays
        k_xz[i] = kxz
        k_yz[i] = kyz
        k_z[i] = kz

    return k_xz, k_yz, k_z



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides classes for analyzing spatially embedded complex networks, handling
multivariate data and generating time series surrogates.
"""

from typing import Optional

import numpy as np
try:
    from h5netcdf.legacyapi import Dataset
except ImportError:
    try:
        from netCDF4 import Dataset
    except ImportError:
        print("pyunicorn: Packages netCDF4 or h5netcdf could not be loaded. "
              "Some functionality in class Data might not be available!")

from .geo_grid import GeoGrid


class Data:

    """
    Encapsulates general spatio-temporal data.

    Also contains methods to load data from various file formats
    (currently NetCDF and ASCII).

    Mainly an abstract class.
    """

    #
    #  Define internal methods
    #

    def __init__(self, observable: np.ndarray, grid: GeoGrid,
                 observable_name: str = None, observable_long_name: str = None,
                 window: Optional[dict] = None, silence_level: int = 0):
        """
        Initialize an instance of Data.

        The spatio-temporal window is described by the following dictionary::

            window = {"time_min": 0., "time_max": 0., "lat_min": 0.,
                      "lat_max": 0., "lon_min": 0., "lon_max": 0.}

        :type observable: 2D array [time, index]
        :arg observable: The array of time series to be represented by the
            :class:`Data` instance.
        :type grid: :class:`.GeoGrid` instance
        :arg grid: The GeoGrid representing the spatial coordinates associated
            to the time series and their temporal sampling.
        :arg str observable_name: A short name for the observable.
        :arg str observable_long_name: A long name for the observable.
        :arg dict window: Spatio-temporal window to select a view on the data.
        :arg int silence_level: The inverse level of verbosity of the object.
        """

        self.silence_level = silence_level
        """(int) - The inverse level of verbosity of the object."""
        self._full_observable = observable

        assert isinstance(grid, GeoGrid)
        self._full_grid = grid
        self.grid = None
        """The :class:`.GeoGrid` object associated with the data."""

        self.observable_name = observable_name
        """(str) - The short name of the observable within
                      data file (particularly relevant for NetCDF)."""

        self.observable_long_name = observable_long_name
        """(str) - The long name of the observable within data file."""

        self._observable = None
        """Current spatio-temporal view on the data."""

        self.file_name = ""
        self.file_type = ""
        self.vertical_level = None

        #  Select a spatio-temporal window to look at the data, can later be
        #  changed in run time without having to reload data from a file.
        #  Force the calling of the Data.setWindow method, since child classes
        #  may overwrite this method.
        if window is None:
            Data.set_global_window(self)
        else:
            Data.set_window(self, window)

    def __str__(self):
        """Return a string representation of the object."""
        if self.file_name:
            self.print_data_info()

        return (f"Data: {self.grid.N} grid points, "
                f"{self.grid.n_grid_points} measurements.\n"
                f"Geographical boundaries:\n{self.grid.print_boundaries()}")

    def set_silence_level(self, silence_level):
        """
        Set the silence level.

        Includes dependent objects such as :attr:`grid`.

        :type silence_level: number (int)
        :arg silence_level: The inverse level of verbosity of the object.
        """
        self.silence_level = silence_level
        self.grid.silence_level = silence_level
        self._full_grid.silence_level = silence_level

    #
    #  Methods for creating Data objects and alternative constructors
    #

    @classmethod
    def Load(cls, file_name, observable_name, file_type, dimension_names=None,
             window=None, vertical_level=None, silence_level=0):
        """
        Initialize an instance of Data.

        Supported file types ``file_type`` are:
          - "NetCDF" for regular (rectangular) grids
          - "iNetCDF" for irregular (e.g. geodesic) grids or station data.

        The :index:`spatio-temporal window` is described by the following
        dictionary::

            window = {"time_min": 0., "time_max": 0., "lat_min": 0.,
                      "lat_max": 0., "lon_min": 0., "lon_max": 0.}

        .. note::
            It is assumed that the NetCDF file to be loaded uses the following
            dimension names: lat, lon, time (e.g., as is the case for
            `NCEP/NCAR reanalysis 1 data <http://www.esrl.noaa.
            gov/psd/data/gridded/data.ncep.reanalysis.html>`_). These standard
            dimension names can be modified using the dimension_names argument.
            Alternatively, the standard class constructor :meth:`__init__`
            needs to be used after loading the data manually, e.g., employing
            netcdf4-python or scipy.io.netcdf functionality.

        :arg str file_name: The name of the data file.
        :arg str observable_name: The short name of the observable within data
            file (particularly relevant for NetCDF).
        :arg str file_type: The type of the data file.
        :arg dict dimension_names: The names of the dimensions as used in the
            NetCDF file. Default: {"lat": "lat", "lon": "lon", "time": "time"}
        :arg dict window: Spatio-temporal window to select a view on the data.
        :arg int vertical_level: The vertical level to be extracted from the
            data file. Is ignored for horizontal data sets. If None, the first
            level in the data file is chosen.
        :arg int silence_level: The inverse level of verbosity of the object.
        """
        if dimension_names is None:
            dimension_names = {"lat": "lat", "lon": "lon", "time": "time"}

        # Import data from given file
        res = cls._load_data(file_name, file_type, observable_name,
                             dimension_names, vertical_level)

        # Create instance of Data
        data = cls(observable=res["observable"], grid=res["grid"],
                   observable_name=res["observable_name"],
                   observable_long_name=res["observable_long_name"],
                   window=window, silence_level=silence_level)

        # Set some variables
        data.file_name = file_name
        data.file_type = file_type
        data.vertical_level = vertical_level

        return data

    @staticmethod
    def SmallTestData():
        """
        Return test data set of 6 time series with 10 sampling points each.

        **Example:**

        >>> Data.SmallTestData().observable()
        array([[  0.00000000e+00,   1.00000000e+00,   1.22464680e-16,
                 -1.00000000e+00,  -2.44929360e-16,   1.00000000e+00],
               [  3.09016994e-01,   9.51056516e-01,  -3.09016994e-01,
                 -9.51056516e-01,   3.09016994e-01,   9.51056516e-01],
               [  5.87785252e-01,   8.09016994e-01,  -5.87785252e-01,
                 -8.09016994e-01,   5.87785252e-01,   8.09016994e-01],
               [  8.09016994e-01,   5.87785252e-01,  -8.09016994e-01,
                 -5.87785252e-01,   8.09016994e-01,   5.87785252e-01],
               [  9.51056516e-01,   3.09016994e-01,  -9.51056516e-01,
                 -3.09016994e-01,   9.51056516e-01,   3.09016994e-01],
               [  1.00000000e+00,   1.22464680e-16,  -1.00000000e+00,
                 -2.44929360e-16,   1.00000000e+00,   3.67394040e-16],
               [  9.51056516e-01,  -3.09016994e-01,  -9.51056516e-01,
                  3.09016994e-01,   9.51056516e-01,  -3.09016994e-01],
               [  8.09016994e-01,  -5.87785252e-01,  -8.09016994e-01,
                  5.87785252e-01,   8.09016994e-01,  -5.87785252e-01],
               [  5.87785252e-01,  -8.09016994e-01,  -5.87785252e-01,
                  8.09016994e-01,   5.87785252e-01,  -8.09016994e-01],
               [  3.09016994e-01,  -9.51056516e-01,  -3.09016994e-01,
                  9.51056516e-01,   3.09016994e-01,  -9.51056516e-01]])

        :rtype: Data instance
        :return: a Data instance for testing purposes.
        """
        #  Create time series
        ts = np.zeros((10, 6))

        for i in range(6):
            ts[:, i] = np.sin(np.arange(10) * np.pi / 10. + i * np.pi / 2.)

        return Data(observable=ts, grid=GeoGrid.SmallTestGrid(),
                    silence_level=2)

    #
    #  Defines methods to load data from files and display related information
    #

    @classmethod
    def _get_netcdf_data(cls, file_name, file_type, observable_name,
                         dimension_names, vertical_level=None,
                         silence_level=0):
        """
        Import data from a NetCDF file with a regular and rectangular grid.

        Supported file types ``file_type`` are:
          - "NetCDF" for regular (rectangular) grids
          - "iNetCDF" for irregular (e.g. geodesic) grids or station data

        :arg str file_name: The name of the data file.
        :arg str file_type: The format of the data file.
        :arg str observable_name: The short name of the observable within data
            file (particularly relevant for NetCDF).
        :arg dict dimension_names: The names of the dimensions as used in the
            NetCDF file. E.g., dimension_names = {"lat": "lat", "lon": "lon",
            "time": "time"}.
        :arg int vertical_level: The vertical level to be extracted from the
            data file. Is ignored for horizontal data sets. If None, the first
            level in the data file is chosen.
        :arg int silence_level: The inverse level of verbosity of the object.
        """
        if silence_level <= 1:
            print("Reading NetCDF File and converting data to NumPy array...")

        # Initialize dictionary of results
        res = {}

        # Open netCDF4 file
        f = Dataset(file_name, "r")

        # Create reference to observable
        observable = f.variables[observable_name][:].astype("float32")

        # Get time axis from NetCDF file
        time = f.variables[dimension_names["time"]][:].astype("float32")

        # Get number of dimensions of data
        n_dim = observable.ndim

        # Distinguish between regular and irregular grids
        if file_type == "NetCDF":
            # Create GeoGrid instance
            lat_grid = f.variables[dimension_names["lat"]][:].astype("float32")
            lon_grid = f.variables[dimension_names["lon"]][:].astype("float32")
            res["grid"] = GeoGrid.RegularGrid(time, (lat_grid, lon_grid),
                                              silence_level)

            # If 3D data set (time, lat, lon), select whole data set
            if n_dim == 3:
                res["observable"] = observable.copy()
            # If 4D data set (time, level, lat, lon), select certain vertical
            # level.
            elif n_dim == 4:
                # Handle selected vertical level
                if vertical_level is None:
                    level = 0
                else:
                    level = vertical_level

                res["observable"] = observable[:, level, :, :].copy()
            else:
                print("Regular NetCDF data sets with dimensions other than "
                      "3 (time, lat, lon) or 4 (time, level, lat, lon) are "
                      "not supported by Data class!")

        elif file_type == "iNetCDF":
            # Create GeoGrid instance
            lat_seq = f.variables["grid_center_lat"][:].astype("float32")
            lon_seq = f.variables["grid_center_lon"][:].astype("float32")
            res["grid"] = GeoGrid(time, lat_seq, lon_seq, silence_level)

            # If 2D data set (time, index), select whole data set
            if n_dim == 2:
                res["observable"] = observable.copy()
            # If 3D data set (time, level, index), select certain vertical
            # level.
            elif n_dim == 3:
                # Handle selected vertical level
                if vertical_level is None:
                    level = 0
                else:
                    level = vertical_level

                res["observable"] = observable[:, level, :].copy()
            else:
                print("Irregular NetCDF data sets with dimensions other than "
                      "2 (time, index) or 3 (time, level, index) are not "
                      "supported by Data class!")

        # Get length of raw data time axis
        n_time = res["observable"].shape[0]
        # Reshape observable to comply with the standard shape (time, index)
        res["observable"].shape = (n_time, -1)

        # Get long name of observable
        res["observable_long_name"] = f.variables[observable_name].long_name

        # Store name of observable
        res["observable_name"] = observable_name

        f.close()
        return res

    @classmethod
    def _load_data(cls, file_name, file_type, observable_name,
                   dimension_names, vertical_level=None, silence_level=0):
        """
        Load data into a Numpy array and create a corresponding GeoGrid object.

        Supported file types ``file_type`` are:
          - "NetCDF" for regular (rectangular) grids
          - "iNetCDF" for irregular (e.g. geodesic) grids or station data

        :arg str file_name: The name of the data file.
        :arg str file_type: The format of the data file.
        :arg str observable_name: The short name of the observable within data
            file (particularly relevant for NetCDF).
        :arg dict dimension_names: The names of the dimensions as used in the
            NetCDF file. E.g., dimension_names = {"lat": "lat", "lon": "lon",
            "time": "time"}.
        :arg int vertical_level: The vertical level to be extracted from the
            data file. Is ignored for horizontal data sets. If None, the first
            level in the data file is chosen.
        :arg int silence_level: The inverse level of verbosity of the object.
        """
        if file_type in ["NetCDF", "iNetCDF"]:
            return cls._get_netcdf_data(file_name, file_type, observable_name,
                                        dimension_names, vertical_level,
                                        silence_level)
        else:
            if silence_level <= 1:
                print("This file type can currently not be read "
                      "by pyunicorn.")
            return None

    def print_data_info(self):
        """Print information on the data encapsulated by the Data object."""
        # Open netCDF4 file
        f = Dataset(self.file_name, "r")
        print("Global attributes:")
        for name in f.ncattrs():
            print(name + ":", getattr(f, name))
        print("Variables (size):")
        for name, obj in f.variables.items():
            print(f"{name} ({len(obj)})")
        f.close()

    def observable(self):
        """
        Return the current spatio-temporal view on the data.

        **Example:**

        >>> Data.SmallTestData().observable()[0,:]
        array([  0.00000000e+00,   1.00000000e+00,   1.22464680e-16,
                -1.00000000e+00,  -2.44929360e-16,   1.00000000e+00])

        :rtype: 2D Numpy array [time, space]
        :return: the current spatio-temporal view on the data.
        """
        return self._observable

    #
    #  Defines methods for windowing the data
    #

    def window(self):
        """
        Return the current spatio-temporal window.

        **Examples:**

        >>> Data.SmallTestData().window()["lon_min"]
        2.5

        >>> Data.SmallTestData().window()["lon_max"]
        15.0

        :rtype: dictionary
        :return: the current spatio-temporal window.
        """
        return self.grid.boundaries()

    def set_window(self, window):
        """
        Select a rectangular spatio-temporal region from the data set.

        Create a data array as well as a corresponding GeoGrid object to access
        this window.

        The time axis of the underlying raw data is assumed to be ordered and
        increasing. The latitude and longitude sequences can be arbitrarily
        chosen, i.e., no ordering and no regular grid is required.

        The spatio-temporal window is described by the following dictionary::

           window = {"time_min": 0., "time_max": 0., "lat_min": 0.,
                     "lat_max": 0., "lon_min": 0., "lon_max": 0.}

        If the temporal boundaries are equal, the data's full time range is
        selected. If any of the two corresponding spatial boundaries are
        equal, the data's full spatial extension is included.

        **Example:**

        >>> data = Data.SmallTestData()
        >>> data.set_window(window={
        ...     "time_min": 0., "time_max": 4., "lat_min": 10.,
        ...     "lat_max": 20., "lon_min": 5., "lon_max": 10.})
        >>> data.observable()
        array([[  1.22464680e-16,  -1.00000000e+00],
               [ -3.09016994e-01,  -9.51056516e-01],
               [ -5.87785252e-01,  -8.09016994e-01],
               [ -8.09016994e-01,  -5.87785252e-01],
               [ -9.51056516e-01,  -3.09016994e-01]])

        :type window: dictionary
        :arg window: A spatio-temporal window to select a view on the data.
        """
        # Collect arrays
        full_time = self._full_grid.grid()["time"]
        full_lat_seq = self._full_grid.grid()["lat"]
        full_lon_seq = self._full_grid.grid()["lon"]

        # Get time indices for temporal window boundaries
        if window["time_min"] == window["time_max"]:
            # If boundaries time are equal, use all available time points
            time_indices = np.repeat(True,
                                     self._full_grid.grid_size()["time"])
        else:
            # Get indices for chosen time boundaries
            time_indices = (full_time >= window["time_min"]) & \
                           (full_time <= window["time_max"])

        # Get indices of nodes lying within the prescribed spatial
        # window boundaries
        if (window["lat_min"] == window["lat_max"]) \
           or (window["lon_min"] == window["lon_max"]):
            # If boundaries in latitude or longitude are equal, use all nodes
            # from the full spatial grid.
            # space_indices is an array of bool indicating whether a node
            # lies within the window or not.
            space_indices = np.repeat(True,
                                      self._full_grid.grid_size()["space"])
        else:
            # space_indices is an array of bool indicating whether a node
            # lies within the window or not.
            space_indices = (full_lat_seq >= window["lat_min"]) & \
                            (full_lat_seq <= window["lat_max"]) & \
                            (full_lon_seq >= window["lon_min"]) & \
                            (full_lon_seq <= window["lon_max"])

        # Set windowed observable and grid object
        time = full_time[time_indices]
        lat_seq = full_lat_seq[space_indices]
        lon_seq = full_lon_seq[space_indices]

        self._observable = \
            self._full_observable[time_indices, :][:, space_indices]
        self.grid = GeoGrid(time, lat_seq, lon_seq, self.silence_level)

    def set_global_window(self):
        """
        Set the view on the whole data set.

        Select the full data set and creates a data array as well as
        a corresponding GeoGrid object to access this window from outside.

        **Example** (Set smaller window and subsequently restore global
        window):

        >>> data = Data.SmallTestData()
        >>> data.set_window(window={"time_min": 0., "time_max": 4.,
        ...                 "lat_min": 10., "lat_max": 20., "lon_min": 5.,
        ...                 "lon_max": 10.})
        >>> data.grid.grid()["lat"]
        array([ 10.,  15.], dtype=float32)
        >>> data.set_global_window()
        >>> data.grid.grid()["lat"]
        array([  0.,   5.,  10.,  15.,  20.,  25.], dtype=float32)
        """
        global_window = {"time_min": 0., "time_max": 0., "lat_min": 0.,
                         "lat_max": 0., "lon_min": 0., "lon_max": 0.}

        self.set_window(global_window)

    #
    # Define methods for reformatting data
    #

    # TODO: Documentation.
    @staticmethod
    def rescale(array, var_type):
        """
        Rescale an array to a given data type.

        Returns the tuple
        (scaled_array, scale_factor, add_offset, actual_range).
        Allows flexible handling of final amount of
        used storage volume for the file.

        :type array:
        :arg array:

        :arg str var_type: Determines the desired final data type of the array.
        """
        # TODO: Add example

        ar_max = array.max()
        ar_min = array.min()
        actual_range = np.array([ar_min, ar_max])

        if var_type == 'float64':
            scaled_array = array.astype('float64')
            scale_factor = 1.
            add_offset = 0.
        elif var_type == 'float32':
            scaled_array = array.astype('float32')
            scale_factor = 1.
            add_offset = 0.
        elif var_type == 'int32':
            scale_factor = (ar_max - ar_min) / (2. * 2. ** 31 - 2.)
            add_offset = (ar_max + ar_min) / 2.
            array -= add_offset
            array /= scale_factor
            scaled_array = array.astype("int32")
        elif var_type == 'int16':
            scale_factor = (ar_max - ar_min) / (2. * 2. ** 15 - 2.)
            add_offset = (ar_max + ar_min) / 2.
            array -= add_offset
            array /= scale_factor
            scaled_array = array.astype('int16')
        elif var_type == 'uint8':
            scale_factor = (ar_max - ar_min) / (2. ** 8 - 1.)
            add_offset = ar_min
            array -= add_offset
            array /= scale_factor
            scaled_array = array.astype('uint8')
        else:
            print(f"Data type {var_type} not supported!")
            scale_factor = 1.
            add_offset = 0.

        return (scaled_array, scale_factor, add_offset, actual_range)

    #
    #  Define methods to prepare data for similarity measure calculation
    #

    @staticmethod
    def normalize_time_series_array(time_series_array):
        """
        :index:`Normalize <pair: normalize; time series array>` an array of
        time series to zero mean and unit variance individually for each
        individual time series.

        Works also for complex valued time series.

        **Modifies the given array in place!**

        **Example:**

        >>> ts = np.arange(16).reshape(4,4).astype("float")
        >>> Data.normalize_time_series_array(ts)
        >>> ts.mean(axis=0)
        array([ 0.,  0.,  0.,  0.])
        >>> ts.std(axis=0)
        array([ 1.,  1.,  1.,  1.])
        >>> ts[:,0]
        array([-1.34164079, -0.4472136 ,  0.4472136 ,  1.34164079])

        :type time_series_array: 2D Numpy array [time, index]
        :arg time_series_array: The time series array to be normalized.
        """
        #  Remove mean value from time series at each node (grid point)
        time_series_array -= time_series_array.mean(axis=0)

        #  Normalize the variance of anomalies to one
        time_series_array /= np.sqrt(
            (time_series_array * time_series_array.conjugate()).mean(axis=0))

        #  Correct for grid points with zero variance in their time series
        time_series_array[np.isnan(time_series_array)] = 0

    @staticmethod
    def next_power_2(i):
        """
        Return the power of two 2^n, that is greater or equal than i.

        **Example:**

        >>> Data.next_power_2(253)
        256

        :type i: number (float)
        :arg i: Some real number.

        :rtype: number (float)
        :return: the power of two greater of equal than a given value.
        """
        n = 2
        while n < i:
            n = n * 2

        return n

    @staticmethod
    def zero_pad_data(data):
        """
        Return :index:`zero padded data`, such that the length of individual
        time series is a power of 2.

        **Example:**

        >>> ts = np.arange(20).reshape(5,4)
        >>> Data.zero_pad_data(ts)
        array([[  0.,   0.,   0.,   0.], [  0.,   1.,   2.,   3.],
               [  4.,   5.,   6.,   7.], [  8.,   9.,  10.,  11.],
               [ 12.,  13.,  14.,  15.], [ 16.,  17.,  18.,  19.],
               [  0.,   0.,   0.,   0.], [  0.,   0.,   0.,   0.]])

        :type data: 2D Numpy array [time, index]
        :arg data: The data array to be zero padded.

        :rtype: 2D Numpy array [time, index]
        :return: the zero padded data array.
        """
        (n_time, n_nodes) = data.shape

        #  Get the power of n that is larger or equal than the length of
        #  individual time series.
        n = Data.next_power_2(n_time)

        zeros_before = (n - n_time) // 2
        zeros_after = (n - n_time) - (n - n_time) // 2

        before = np.zeros((zeros_before, n_nodes))
        after = np.zeros((zeros_after, n_nodes))

        return np.concatenate((before, data, after), axis=0)

    @staticmethod
    def cos_window(data, gamma):
        """
        Return a cosine window fitting the shape of the data argument.

        The window is one for most of the time and goes to zero at the
        boundaries of each time series in the data array.

        The width of the cosine shaped decay region is controlled by the shape
        parameter gamma:

          - Gamma=1 means, that each of the two decay regions extends over
            half of the time series.
          - Gamma=0 means, that the decay regions vanish and the window
            transformation becomes the identity.

        **Example:**

        >>> ts = np.arange(24).reshape(12,2)
        >>> Data.cos_window(data=ts, gamma=0.75)
        array([[ 0.        ,  0.        ], [ 0.14644661,  0.14644661],
               [ 0.5       ,  0.5       ], [ 0.85355339,  0.85355339],
               [ 1.        ,  1.        ], [ 1.        ,  1.        ],
               [ 1.        ,  1.        ], [ 1.        ,  1.        ],
               [ 0.85355339,  0.85355339], [ 0.5       ,  0.5       ],
               [ 0.14644661,  0.14644661], [ 0.        ,  0.        ]])

        :type data: 2D Numpy array [time, index]
        :arg data: The data array to be fitted by cosine window.

        :type gamma: number (float)
        :arg gamma: The cosine window shape parameter.

        :rtype: 2D Numpy array [time, index]
        :return: the cosine window fitting data array.
        """
        (n_time, n_nodes) = data.shape

        #  Calculate length of decay regions
        decay_length = int(gamma * n_time / 2)

        #  Calculate decay and growth regions
        growth_region = 0.5 * (1 + np.cos(
            np.arange(decay_length) * np.pi / float(decay_length) + np.pi))
        growth_region = np.tile(growth_region, (n_nodes, 1))
        growth_region = growth_region.transpose()

        decay_region = growth_region[::-1, :].copy()

        return np.concatenate((growth_region,
                               np.ones((n_time - 2 * decay_length, n_nodes)),
                               decay_region), axis=0)



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides classes for analyzing spatially embedded complex networks, handling
multivariate data and generating time series surrogates.
"""

# array object and fast numerics
import numpy as np
from numpy import random

# easy progress bar handling
from tqdm import trange

from ..core._ext.types import to_cy, ADJ, DEGREE, FIELD, DFIELD
from ._ext.numerics import _embed_time_series_array, _recurrence_plot, \
    _twins_s, _twin_surrogates_s, _test_pearson_correlation, \
    _test_mutual_information

#
#  Define class Surrogates
#


class Surrogates:

    """
    Encapsulates structures and methods related to surrogate time series.

    Provides data structures and methods to generate surrogate data sets from a
    set of time series and to evaluate the significance of various correlation
    measures using these surrogates.

    More information on time series surrogates can be found in [Schreiber2000]_
    and [Kantz2006]_.
    """

    #
    #  Define internal methods
    #
    def __init__(self, original_data, silence_level=1):
        """
        Initialize an instance of Surrogates.

        .. note::
           The order of array dimensions is different from the standard of
           ``core``. Here it is [index, time] for reasons of computational
           speed!

        :type original_data: 2D array [index, time]
        :arg original_data: The original time series for surrogate generation.
        :arg int silence_level: The inverse level of verbosity of the object.
        """
        if silence_level <= 1:
            print("Generated an instance of the Surrogates class.")

        #  Set class variables
        self.original_data = original_data
        """The original time series for surrogate generation."""
        self.silence_level = silence_level
        """(string) - The inverse level of verbosity of the object."""

        #  Set flags
        self._normalized = False
        self._fft_cached = False
        self._twins_cached = False

        #  Cache
        self._twins = None
        self._original_data_fft = None

    def __str__(self):
        """
        Returns a string representation.
        """
        return f"Surrogates: time series shape {self.original_data.shape}."

    def clear_cache(self):
        """Clean up cache."""
        try:
            del self._original_data_fft
            del self._twins
        except AttributeError:
            pass

    #
    #  Methods for testing purposes
    #

    @staticmethod
    def SmallTestData():
        """
        Return Surrogates instance representing test a data set of 6 time
        series.

        :rtype: Surrogates instance
        :return: a Surrogates instance for testing purposes.
        """
        #  Create time series
        ts = np.zeros((6, 200))

        for i in range(6):
            ts[i, :] = np.sin(np.arange(200)*np.pi/15. + i*np.pi/2.) + \
                np.sin(np.arange(200) * np.pi / 30.)

        return Surrogates(original_data=ts, silence_level=2)

    #
    #  Define methods to normalize and analyze the data
    #

    @staticmethod
    def normalize_time_series_array(time_series_array):
        """
        :index:`Normalize <pair: normalize; time series array>` an array of
        time series to zero mean and unit variance individually for each
        individual time series.

        **Modifies the given array in place!**

        **Examples:**

        >>> ts = Surrogates.SmallTestData().original_data
        >>> Surrogates.SmallTestData().normalize_time_series_array(ts)
        >>> r(ts.mean(axis=1))
        array([ 0., 0., 0., 0., 0., 0.])
        >>> r(ts.std(axis=1))
        array([ 1., 1., 1., 1., 1., 1.])

        :type time_series_array: 2D array [index, time]
        :arg time_series_array: The time series array to be normalized.
        """
        mean = time_series_array.mean(axis=1)
        std = time_series_array.std(axis=1)

        for i in range(time_series_array.shape[0]):
            #  Remove mean value from time series at each node (grid point)
            time_series_array[i, :] -= mean[i]
            #  Normalize the standard deviation of anomalies to one
            if std[i] != 0:
                time_series_array[i, :] /= std[i]

    def embed_time_series_array(self, time_series_array, dimension, delay):
        """
        Return a :index:`delay embedding` of all time series.

        .. note::
           Only works for scalar time series!

        **Example:**

        >>> ts = Surrogates.SmallTestData().original_data
        >>> Surrogates.SmallTestData().embed_time_series_array(
        ...     time_series_array=ts, dimension=3, delay=2)[0,:6,:]
        array([[ 0.        ,  0.61464833,  1.14988147],
               [ 0.31244015,  0.89680225,  1.3660254 ],
               [ 0.61464833,  1.14988147,  1.53884177],
               [ 0.89680225,  1.3660254 ,  1.6636525 ],
               [ 1.14988147,  1.53884177,  1.73766672],
               [ 1.3660254 ,  1.6636525 ,  1.76007351]])

        :type time_series_array: 2D array [index, time]
        :arg time_series_array: The time series array to be normalized.
        :arg int dimension: The embedding dimension.
        :arg int delay: The embedding delay.
        :rtype: 3D array [index, time, dimension]
        :return: the embedded time series.
        """
        if self.silence_level <= 1:
            print(f"Embedding all time series in dimension {dimension} "
                  f"and with lag {delay} ...")
        (N, n_time) = time_series_array.shape

        embedding = np.empty(
            (N, n_time - (dimension - 1)*delay, dimension), dtype=DFIELD)

        _embed_time_series_array(
            N, n_time, dimension, delay,
            to_cy(time_series_array, DFIELD), embedding)
        return embedding

    # FIXME: I(wb) included the line
    # dimension = embedding.shape[1]
    # whose missing caused an error. I can't guarantee if it is correct.
    def recurrence_plot(self, embedding, threshold):
        """
        Return the :index:`recurrence plot <pair: recurrence plot; time
        series>` from an embedding of a time series.

        Uses supremum norm.

        :type embedding: 2D array [time, dimension]
        :arg embedding: The embedded time series.
        :arg float threshold: The recurrence threshold.
        :rtype: 2D array [time, time]
        :return: the recurrence matrix.
        """
        if self.silence_level <= 1:
            print("Calculating the recurrence plot...")

        n_time = embedding.shape[0]
        dimension = embedding.shape[1]
        R = np.ones((n_time, n_time), dtype=ADJ)

        _recurrence_plot(n_time, dimension, threshold,
                         to_cy(embedding, FIELD), R)
        return R

    # FIXME: I(wb) included the line
    # dimension = embedding_array.shape[2]
    # whose missing caused an error. I can't guarantee if it is correct.
    def twins(self, embedding_array, threshold, min_dist=7):
        """
        Return list of the :index:`twins <pair: twins; surrogates>` of each
        state vector for all time series.

        Two state vectors are said to be twins if they share the same
        recurrences, i.e., if the corresponding rows or columns in the
        recurrence plot are identical.

        References: [Thiel2006]_, [Marwan2007]_.

        :type embedding_array: 3D array [index, time, dimension]
        :arg  embedding_array: The embedded time series array.
        :arg float threshold: The recurrence threshold.
        :arg number min_dist: The minimum temporal distance for twins.
        :rtype: [[number]]
        :return: the list of twins for each state vector in the time series.
        """
        if self.silence_level <= 1:
            print("Finding twins...")

        N = embedding_array.shape[0]
        n_time = embedding_array.shape[1]
        dimension = embedding_array.shape[2]
        twins = []

        #  Initialize the R matrix with ones
        R = np.empty((n_time, n_time), dtype=ADJ)
        #  Initialize array to store the number of neighbors for each sample
        nR = np.empty(n_time, dtype=DEGREE)

        _twins_s(N, n_time, dimension, threshold, min_dist,
                 to_cy(embedding_array, DFIELD), R, nR, twins)

        return twins

    #
    #  Define methods to generate sets of surrogate time series
    #

    def white_noise_surrogates(self, original_data):
        """
        Return a shuffled copy of a time series array.

        Each time series is shuffled individually. The surrogates correspond to
        realizations of white noise consistent with the :attr:`original_data`
        time series' amplitude distribution.

        **Example** (Distributions of white noise surrogates should the same as
        for the original data):

        >>> ts = Surrogates.SmallTestData().original_data
        >>> surrogates = Surrogates.\
                SmallTestData().white_noise_surrogates(ts)
        >>> np.allclose(np.histogram(ts[0,:])[0],
        ...             np.histogram(surrogates[0,:])[0])
        True

        :type original_data: 2D array [index, time]
        :arg original_data: The original time series.
        :rtype: 2D array [index, time]
        :return: The surrogate time series.
        """
        if self.silence_level <= 1:
            print("Generating white noise surrogates by random shuffling...")

        #  Generate reference to shuffle function
        shuffle = random.shuffle

        surrogates = original_data.copy()

        for i in range(surrogates.shape[0]):
            shuffle(surrogates[i, :])

        return surrogates

    def correlated_noise_surrogates(self, original_data):
        """
        Return Fourier surrogates.

        Generate surrogates by Fourier transforming the :attr:`original_data`
        time series (assumed to be real valued), randomizing the phases and
        then applying an inverse Fourier transform. Correlated noise surrogates
        share their power spectrum and autocorrelation function with the
        original_data time series.

        The Fast Fourier transforms of all time series are cached to facilitate
        a faster generation of several surrogates for each time series. Hence,
        :meth:`clear_cache` has to be called before generating surrogates from
        a different set of time series!

        .. note::
           The amplitudes are not adjusted here, i.e., the
           individual amplitude distributions are not conserved!

        **Examples:**

        The power spectrum is conserved up to small numerical deviations:

        >>> ts = Surrogates.SmallTestData().original_data
        >>> surrogates = Surrogates.\
                SmallTestData().correlated_noise_surrogates(ts)
        >>> all(r(np.abs(np.fft.fft(ts,         axis=1))[0,1:10]) == \
                r(np.abs(np.fft.fft(surrogates, axis=1))[0,1:10]))
        True

        However, the time series amplitude distributions differ:

        >>> all(np.histogram(ts[0,:])[0] == np.histogram(surrogates[0,:])[0])
        False

        :type original_data: 2D array [index, time]
        :arg original_data: The original time series.
        :rtype: 2D array [index, time]
        :return: The surrogate time series.
        """
        if self.silence_level <= 1:
            print("Generating correlated noise surrogates...")

        #  Calculate FFT of original_data time series
        #  The FFT of the original_data data has to be calculated only once,
        #  so it is stored in self._original_data_fft.
        if self._fft_cached:
            surrogates = self._original_data_fft
        else:
            surrogates = np.fft.rfft(original_data, axis=1)
            self._original_data_fft = surrogates
            self._fft_cached = True

        #  Get shapes
        (N, n_time) = original_data.shape
        len_phase = surrogates.shape[1]

        #  Generate random phases uniformly distributed in the
        #  interval [0, 2*Pi]
        phases = random.uniform(low=0, high=2 * np.pi, size=(N, len_phase))

        #  Add random phases uniformly distributed in the interval [0, 2*Pi]
        surrogates *= np.exp(1j * phases)

        #  Calculate IFFT and take the real part, the remaining imaginary part
        #  is due to numerical errors.
        return np.ascontiguousarray(np.real(np.fft.irfft(surrogates, n=n_time,
                                                         axis=1)))

    def AAFT_surrogates(self, original_data):
        """
        Return surrogates using the amplitude adjusted Fourier transform
        method.

        Reference: [Schreiber2000]_

        :type original_data: 2D array [index, time]
        :arg original_data: The original time series.
        :rtype: 2D array [index, time]
        :return: The surrogate time series.
        """
        #  Create sorted Gaussian reference series
        gaussian = random.randn(original_data.shape[0], original_data.shape[1])
        gaussian.sort(axis=1)

        #  Rescale data to Gaussian distribution
        ranks = original_data.argsort(axis=1).argsort(axis=1)
        rescaled_data = np.zeros(original_data.shape)

        for i in range(original_data.shape[0]):
            rescaled_data[i, :] = gaussian[i, ranks[i, :]]

        #  Phase randomize rescaled data
        phase_randomized_data = \
            self.correlated_noise_surrogates(rescaled_data)

        #  Rescale back to amplitude distribution of original data
        sorted_original = original_data.copy()
        sorted_original.sort(axis=1)

        ranks = phase_randomized_data.argsort(axis=1).argsort(axis=1)

        for i in range(original_data.shape[0]):
            rescaled_data[i, :] = sorted_original[i, ranks[i, :]]

        return rescaled_data

    def refined_AAFT_surrogates(self, original_data, n_iterations,
                                output="true_amplitudes"):
        """
        Return surrogates using the iteratively refined amplitude adjusted
        Fourier transform method.

        A set of AAFT surrogates (:meth:`AAFT_surrogates`) is iteratively
        refined to produce a closer match of both amplitude distribution and
        power spectrum of surrogate and original data.

        Reference: [Schreiber2000]_

        :type original_data: 2D array [index, time]
        :arg original_data: The original time series.
        :arg int n_iterations: Number of iterations / refinement steps
        :arg str output: Type of surrogate to return. "true_amplitudes":
            surrogates with correct amplitude distribution, "true_spectrum":
            surrogates with correct power spectrum, "both": return both outputs
            of the algorithm.
        :rtype: 2D array [index, time]
        :return: The surrogate time series.
        """
        #  Get size of dimensions
        n_time = original_data.shape[1]

        #  Get Fourier transform of original data with caching
        if self._fft_cached:
            fourier_transform = self._original_data_fft
        else:
            fourier_transform = np.fft.rfft(original_data, axis=1)
            self._original_data_fft = fourier_transform
            self._fft_cached = True

        #  Get Fourier amplitudes
        original_fourier_amps = np.abs(fourier_transform)

        #  Get sorted copy of original data
        sorted_original = original_data.copy()
        sorted_original.sort(axis=1)

        #  Get starting point / initial conditions for R surrogates
        # (see [Schreiber2000]_)
        R = self.AAFT_surrogates(original_data)

        #  Start iteration
        for i in range(n_iterations):
            #  Get Fourier phases of R surrogate
            r_fft = np.fft.rfft(R, axis=1)
            r_phases = r_fft / np.abs(r_fft)

            #  Transform back, replacing the actual amplitudes by the desired
            #  ones, but keeping the phases exp(iψ(i)
            s = np.fft.irfft(original_fourier_amps * r_phases, n=n_time,
                             axis=1)

            #  Rescale to desired amplitude distribution
            ranks = s.argsort(axis=1).argsort(axis=1)

            for j in range(original_data.shape[0]):
                R[j, :] = sorted_original[j, ranks[j, :]]

        if output == "true_amplitudes":
            return R
        elif output == "true_spectrum":
            return s
        elif output == "both":
            return (R, s)
        else:
            return (R, s)

    def twin_surrogates(self, original_data, dimension, delay, threshold,
                        min_dist=7):
        """
        Return surrogates using the twin surrogate method.

        Scalar twin surrogates are created by isolating the first component
        (dimension) of the twin surrogate trajectories.

        Twin surrogates share linear and nonlinear properties with the original
        time series, since they correspond to realizations of trajectories of
        the same dynamical systems with different initial conditions.

        References: [Thiel2006]_ [*], [Marwan2007]_.

        The twin lists of all time series are cached to facilitate a faster
        generation of several surrogates for each time series. Hence,
        :meth:`clear_cache` has to be called before generating twin surrogates
        from a different set of time series!

        :type original_data: 2D array [index, time]
        :arg original_data: The original time series.
        :arg int dimension: The embedding dimension.
        :arg int delay: The embedding delay.
        :arg float threshold: The recurrence threshold.
        :arg number min_dist: The minimum temporal distance for twins.
        :rtype: 2D array [index, time]
        :return: the twin surrogates.
        """
        #  The algorithm proceeds in several steps:
        #  1. Embed the original_data time series, using time delay embedding
        #     for simplicity. Use the same dimension and time delay delay for
        #     all time series for simplicity. Determine delay using time
        #     delayed mutual information and d using false nearest neighbors
        #     methods.
        #  2. Use the algorithm proposed in [*] to find twins
        #  3. Reconstruct one-dimensional twin surrogate time series

        (N, n_time) = original_data.shape
        n_time = n_time - (dimension-1)*delay

        #  Make sure that twins are calculated only once
        if self._twins_cached:
            twins = self._twins
        else:
            embedding = self.embed_time_series_array(original_data,
                                                     dimension, delay)
            twins = self.twins(embedding, threshold, min_dist)
            self._twins = twins
            self._twins_cached = True

        return _twin_surrogates_s(N, n_time, twins,
                                  to_cy(original_data, DFIELD))

    #
    #  Defines methods to generate correlation measure matrices based on
    #  original_data and surrogate data for significance testing.
    #

    @staticmethod
    def eval_fast_code(function, original_data, surrogates):
        """
        Evaluate performance of fast and slow versions of algorithms.

        Designed for evaluating fast and dirty C code against cleaner code
        using Blitz arrays. Does some profiling and returns the total error
        between the results.

        :type function: Python function
        :arg function: The function to be evaluated.
        :type original_data: 2D array [index, time]
        :arg original_data: The original time series.
        :type surrogates: 2D array [index, time]
        :arg surrogates: The surrogate time series.
        :return float: The total squared difference between resulting matrices.
        """
        #  Some profiling
        # profile.run("fastResult = function(original_data, surrogates,
        #             fast=True)")
        # profile.run("slowResult = function(original_data, surrogates,
        #             fast=False)")

        fast_result = function(original_data, surrogates, fast=True)
        slow_result = function(original_data, surrogates, fast=False)

        #  Return error
        return np.sqrt(((fast_result - slow_result)**2).sum())

    @staticmethod
    def test_pearson_correlation(original_data, surrogates):
        """
        Return a test matrix of the Pearson correlation coefficient (zero lag).

        The test matrix's entry :math:`(i,j)` contains the Pearson correlation
        coefficient between original time series i and surrogate time series j
        at lag zero. The resulting matrix is useful for significance tests
        based on the Pearson correlation matrix of the original data.

        .. note::
           Assumes, that original_data and surrogates are already normalized.

        :type original_data: 2D array [index, time]
        :arg original_data: The original time series.
        :type surrogates: 2D array [index, time]
        :arg surrogates: The surrogate time series.
        :rtype: 2D array [index, index]
        :return: the Pearson correlation test matrix.
        """
        (N, n_time) = original_data.shape
        return _test_pearson_correlation(to_cy(original_data, DFIELD),
                                         to_cy(surrogates, DFIELD),
                                         N, n_time)

    @staticmethod
    def test_mutual_information(original_data, surrogates, n_bins=32):
        """
        Return a test matrix of mutual information (zero lag).

        The test matrix's entry :math:`(i,j)` contains the mutual information
        between original time series i and surrogate time series j at zero lag.
        The resulting matrix is useful for significance tests based on the
        mutual information matrix of the original data.

        .. note::
           Assumes, that original_data and surrogates are already normalized.

        :type original_data: 2D array [index, time]
        :arg original_data: The original time series.
        :type surrogates: 2D Numpy array [index, time]
        :arg surrogates: The surrogate time series.
        :arg int n_bins: Number of bins for estimating prob. distributions.
        :rtype: 2D array [index, index]
        :return: the mutual information test matrix.
        """
        (N, n_time) = original_data.shape
        #  Calculate symbolic time series and histograms
        #  Calculate 2D histograms and mutual information
        #  mi[i,j] gives the mutual information between the ith original_data
        #  time series and the jth surrogate time series.
        return _test_mutual_information(to_cy(original_data, DFIELD),
                                        to_cy(surrogates, DFIELD),
                                        N, n_time, n_bins)

    #
    #  Define methods to perform significance tests on correlation measures
    #  based on surrogates.
    #

    def original_distribution(self, test_function, original_data, n_bins=100):
        """
        Return a normalized histogram of a similarity measure matrix.

        The absolute value of the similarity measure is used, since only the
        degree of similarity was of interest originally.

        :type test_function: Python function
        :arg test_function: The function implementing the similarity measure.
        :type original_data: 2D array [index, time]
        :arg original_data: The original time series.
        :arg int n_bins: The number of bins for estimating prob. distributions.
        :rtype: tuple of 1D arrays ([bins],[bins])
        :return: the similarity measure histogram and lower bin boundaries.
        """
        if self.silence_level <= 1:
            print("Estimating probability density distribution of "
                  "original_data ...")

        #  Normalize original_data time series to zero mean and unit variance
        if not self._normalized:
            self.normalize_time_series_array(original_data)
            self._normalized = True

        correlation_measure = np.abs(test_function(original_data,
                                                   original_data))
        (hist, lbb) = np.histogram(correlation_measure, n_bins, density=True)
        #  Normalize
        hist /= hist.sum()

        lbb = lbb[:-1]

        return (hist, lbb)

    def test_threshold_significance(self, surrogate_function, test_function,
                                    realizations=1, n_bins=100,
                                    interval=(-1, 1)):
        """
        Return a test distribution for a similarity measure.

        Perform a significance test on the values of a correlation measure
        based on original_data time series and surrogate data. Returns a
        density estimate (histogram) of the absolute value of the correlation
        measure over all realizations.

        The resulting distribution of the values of similarity measure from
        original and surrogate time series is of use for testing the
        statistical significance of a selected threshold value for climate
        network generation.

        :type surrogate_function: Python function
        :arg surrogate_function: The function implementing the surrogates.
        :type test_function: Python function
        :arg test_function: The function implementing the similarity measure.
        :arg int realizations: The number of surrogates to be created for each
            time series.
        :arg int n_bins: The number of bins for estimating probability
            distribution of test similarity measure.
        :type interval: (float, float)
        :arg interval: The range over which to estimate similarity measure
            distribution.
        :rtype: tuple of 1D arrays ([bins],[bins])
        :return: similarity measure test histogram and lower bin boundaries.
        """
        if self.silence_level <= 1:
            print(f"Starting significance test based on {realizations} "
                  "realizations of surrogates...")

        original_data = self.original_data
        self._fft_cached = False
        self._twins_cached = False

        #  Create reference to np.histogram function
        numpy_hist = np.histogram

        #  Normalize original_data time series to zero mean and unit variance
        if not self._normalized:
            self.normalize_time_series_array(original_data)
            self._normalized = True

        #  Initialize density estimate
        density_estimate = np.zeros(n_bins)

        for _ in trange(realizations, disable=self.silence_level > 2):
            #  Get the surrogate
            #  Mean and variance are conserved by all surrogates
            surrogates = surrogate_function(original_data)

            #  Get the correlation measure test matrix
            correlation_measure_test = np.abs(test_function(original_data,
                                                            surrogates))

            #  Test if correlation measure values are outside range
            if correlation_measure_test.min() < interval[0]:
                print("Warning! Correlation measure value left of range.")
            if correlation_measure_test.max() > interval[1]:
                print("Warning! Correlation measure value right of range.")

            #  Estimate density of current realization
            (hist, lbb) = numpy_hist(correlation_measure_test, n_bins,
                                     interval, normed=True)

            #  Add to density estimate over all realizations
            density_estimate += hist

            #  Clean up (should be done automatically by Python,
            #  but you never know...)
            del surrogates, correlation_measure_test

        #  Normalize density estimate
        density_estimate /= density_estimate.sum()

        lbb = lbb[:-1]

        return (density_estimate, lbb)



# Copyright (C) 2014 SWIPO Project
#
# Authors (this file):
#   Stefan Schinkel <stefan.schinkel@gmail.com>
"""
Weave tests to check that python and weave implementations give the same
results
"""

import numpy as np
from pyunicorn import ResNetwork


res = ResNetwork.SmallTestNetwork()
resC = ResNetwork.SmallComplexNetwork()


def testVCFB():
    for i in range(5):
        admittance = res.get_admittance()
        R = res.get_R()
        # set params
        Is = It = np.float32(1.0)
        # alloc output
        vcfbPython = np.float32(0)
        for t in range(res.N):
            for s in range(t):
                K = 0.0
                if i in (t, s):
                    pass
                else:
                    for j in range(res.N):
                        K += admittance[i][j] * np.abs(
                            Is*(R[i][s]-R[j][s]) + It*(R[j][t]-R[i][t]))/2.
                vcfbPython += 2.*K/(res.N*(res.N-1))

        vcfbCython = res.vertex_current_flow_betweenness(i)
        assert round(vcfbPython, 4) == round(vcfbCython, 4)


def testECFB():
    # set currents
    Is = It = np.float32(1)
    # alloc output
    if res.flagComplex:
        dtype = complex
    else:
        dtype = float
    ecfbPython = np.zeros([res.N, res.N], dtype=dtype)
    # the usual
    admittance = res.get_admittance()
    R = res.get_R()
    for i in range(res.N):
        for j in range(res.N):
            K = 0
            for t in range(res.N):
                for s in range(t):
                    K += admittance[i][j] * np.abs(
                        Is*(R[i][s]-R[j][s])+It*(R[j][t]-R[i][t]))
            # Lets try to compute the in
            ecfbPython[i][j] = 2*K/(res.N*(res.N-1))

    ecfbCython = res.edge_current_flow_betweenness()
    L = len(ecfbPython)
    for i in range(L):
        for j in range(L):
            assert round((ecfbPython[i][j].astype('float32')), 4) == \
                   round((ecfbCython[i][j].astype('float32')), 4)



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides classes for generating and analyzing complex climate networks.
"""

#
#  Import essential packages
#

# import numpy as np

from ..core._ext.types import to_cy, MASK, FIELD
from ._ext.numerics import spearman_corr

#  Import cnTsonisClimateNetwork for TsonisClimateNetwork class
from .climate_network import ClimateNetwork

#
#  Define class RainfallClimateNetwork
#


class RainfallClimateNetwork(ClimateNetwork):

    """
    Encapsulate a Rainfall climate network.

    The Rainfall climate network is constructed from the Spearman rank order
    correlation matrix (Spearman's rho) but without considering "zeros" in the
    dataset, which represent the time at which there is no rainfall.
    Spearman's rho is more robust with respect to outliers and non-gaussian
    data distributions than the Pearson correlation coefficient.

    Rainfall climate networks are undirected due to the symmetry of the
    Spearman's rho matrix.

    Class RainfallClimateNetwork was created by `Marc Wiedermann
    <marcw@physik.hu-berlin.de>`__ during an internship
    at PIK in March 2010.
    """

    #
    # Defines internal methods
    #

    def __init__(self, data, threshold=None, link_density=None,
                 non_local=False, node_weight_type="surface",
                 event_threshold=(0, 1), scale_fac=37265, offset=10**(-7),
                 silence_level=0):
        """
        Initialize an instance of RainfallClimateNetwork.

        .. note::
           Either threshold **OR** link_density have to be given!

        Possible choices for ``node_weight_type``:
          - None (constant unit weights)
          - "surface" (cos lat)
          - "irrigation" (cos**2 lat)

        :type data: :class:`.ClimateData`
        :arg data: The climate data used for network construction.
        :arg float threshold: The threshold of similarity measure, above which
            two nodes are linked in the network.
        :arg float link_density: The networks's desired link density.
        :arg bool non_local: Determines, whether links between spatially close
            nodes should be suppressed.
        :arg str node_weight_type: The type of geographical node weight to be
            used.
        :type event_threshold: list of two numbers between 0 and 1.
        :arg event_threshold: The quantiles of the rainfall distribution at
            each location between which rainfall events should be considered
            for calculating correlations.
        :arg float scale_fac: Scale factor for rescaling data.
        :arg float offset: Offset for rescaling data.
        :arg int silence_level: The inverse level of verbosity of the object.
        """
        if silence_level <= 1:
            print("Generating a Rainfall climate network...")

        #  Set instance variables
        self.data = data
        """(ClimateData) - The climate data used for network construction."""

        self.N = self.data.grid.N
        self._threshold = threshold
        self._prescribed_link_density = link_density
        self._non_local = non_local
        self.node_weight_type = node_weight_type
        self.silence_level = silence_level
        self.time_cycle = self.data.time_cycle

        #  Calculate correlation measure
        correlation = self._calculate_correlation(
            event_threshold, scale_fac, offset, time_cycle=self.time_cycle)

        ClimateNetwork.__init__(self, grid=self.data.grid,
                                similarity_measure=correlation,
                                threshold=self.threshold(),
                                link_density=self._prescribed_link_density,
                                non_local=self.non_local(),
                                directed=False,
                                node_weight_type=self.node_weight_type,
                                silence_level=self.silence_level)

    def __str__(self):
        """
        Returns a string representation of RainfallClimateNetwork.
        """
        return 'RainfallClimateNetwork:\n' + ClimateNetwork.__str__(self)

    #
    # Defines methods to calculate the correlation matrix
    #

    def _calculate_correlation(self, event_threshold, scale_fac, offset,
                               time_cycle):
        """
        Returns the Spearman Rho correlation matrix.

        An event_threshold can be given to extract a percentage of the given
        dataset, i.e. [0.9,1] extracts the ten percent of heaviest rainfall
        events. [0,1] selects the whole dataset.

        :type event_threshold: list of two numbers between 0 and 1.
        :arg event_threshold: The quantiles of the rainfall distribution at
                              each location between which rainfall events
                              should be considered for calculating
                              correlations.

        :type scale_fac: number (float)
        :arg scale_fac: Scale factor for rescaling data.

        :type offset: number (float)
        :arg offset: Offset for rescaling data.

        :type time_cycle: number (int)
        :arg time_cycle: Length of annual cycle in given data (monthly: 12,
                         daily: 365 etc.)

        :rtype: 2D Numpy array (index, index)
        :return: the Spearman's rho matrix at zero lag.
        """
        # Calculate the real rainfall from observable
        rainfall = self.calculate_rainfall(self.data.observable().T,
                                           scale_fac, offset)

        if self.silence_level <= 1:
            print("Calculating Rainfall-Anomaly using Cython...")

        # Calculate the anomaly for the rainfall dataset
        anomaly = self.calculate_rainfall(self.data.anomaly().T,
                                          scale_fac, offset)

        # Correct anomaly for offset due to rescaling
        anomaly -= scale_fac * offset

        # Create a mask, which filters out zeros and events outside the
        # event_threshold
        final_mask = self.calculate_top_events(rainfall, event_threshold)

        if self.silence_level <= 1:
            print("Calculating Spearman-Rho-Matrix using Cython...")

        # Return the correlation matrix
        return self.spearman_corr(final_mask, anomaly)

    @staticmethod
    def calculate_rainfall(observable, scale_fac, offset):
        """
        Returns the rainfall in mm on each measuring point.

        :type observable: 2D Numpy array (time, index)
        :arg observable: The observable time series from the data source.

        :type scale_fac: number (float)
        :arg scale_fac: Scale factor for rescaling data.

        :type offset: number (float)
        :arg offset: Offset for rescaling data.

        :rtype: 2D Numpy array (time, index)
        :return: the rainfall for each time and location
        """
        # Multiply the observable with the scaling factor after adding the
        # offset
        rainfall = (observable + offset) * scale_fac

        return rainfall

    @staticmethod
    def calculate_top_events(rainfall, event_threshold):
        """
        Returns a mask with boolean values. The entries are false, when the
        rainfall of one day is zero, or when the rainfall is not inside the
        event_treshold

        :type rainfall: 2D Numpy array (index, time)
        :arg rainfall: the rainfall time series for each measuring point

        :type event_threshold: list of two numbers between 0 and 1.
        :arg event_threshold: The quantiles of the rainfall distribution at
                              each location between which rainfall events
                              should be considered for calculating
                              correlations.

        :rtype: 2D Numpy array (index, time)
        :return: A bool array with False for every value in the rainfall
                 data, which are zero or outside the top_event Interval.
        """
        rainfall_copy = rainfall.copy()

        m = len(rainfall) * len(rainfall.T)

        onelist = rainfall.reshape(m)

        onelist = onelist[onelist.sort()][0]

        downlimit = m * event_threshold[0] // 1

        uplimit = m * event_threshold[1] // 1

        rainfall = rainfall_copy

        down_mask = rainfall >= onelist[downlimit]

        up_mask = rainfall <= onelist[uplimit-1]

        no_rain_mask = rainfall != 0

        final_mask = down_mask & up_mask & no_rain_mask

        return final_mask

    @staticmethod
    def rank_time_series(anomaly):
        """
        Return rank time series.

        :type anomaly: 2D Numpy array (index, time)
        :arg anomaly: the rainfall anomaly time series for each measuring point

        :rtype: 2D Numpy array (index, time)
        :return: The ranked time series for each gridpoint
        """
        rank_time_series = anomaly.argsort(axis=1).argsort(axis=1) + 1.0
        return rank_time_series

    def spearman_corr(self, final_mask, anomaly):
        """
        Return the Spearman Correlation Matrix at zero lag.

        :type final_mask: 2D Numpy array (index, time)
        :arg final_mask: A bool array with False for every value in the
                         rainfall data, which are zero or outside the top_event
                         interval.

        :type anomaly: 2D Numpy array (index, time)
        :arg anomaly: The rainfall anomaly time series for each measuring
                      point.

        :rtype: 2D Numpy array (index, index)
        :return: the Spearman correlation matrix.
        """
        # Get rank time series
        time_series_ranked = self.rank_time_series(anomaly)
        m, tmax = anomaly.shape
        return spearman_corr(
            m, tmax, to_cy(final_mask, MASK), to_cy(time_series_ranked, FIELD))



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
timeseries
==========

Provides classes for the analysis of dynamical systems and time series based on
recurrence plots, including measures of recurrence quantification analysis
(RQA) and recurrence network analysis.

Also contains methods to calculate the RQA (recurrence quantification analysis)
measures of the recurrence plot.

Related Publications
~~~~~~~~~~~~~~~~~~~~
Recurrence plots, RQA, Recurrence Network Analysis
--------------------------------------------------
[Marwan2007]_, [Marwan2009]_, [Donner2010a]_, [Donner2010b]_, [Donner2011a]_,
[Donner2011b]_, [Zou2010]_, [Donges2011b]_, [Donges2011c]_, [Donges2012]_,
[Feldhoff2012]_, [Zou2012]_, [Feldhoff2013]_

Visibility Graph Analysis of Geoscientific Time Series
------------------------------------------------------
[Donner2012]_

To do
~~~~~
  - Implement access to threshold_from_recurrence_rate_fast from
    constructor of RecurrencePlot.
  - Add more RQA measures and improve their documentation, including formulas.
  - Add random node relabeling to Network class.

Known Bugs
~~~~~~~~~~
  - RQA measures are not yet calculated correctly for Eckmann recurrence plots.

"""

from .cross_recurrence_plot import CrossRecurrencePlot
from .inter_system_recurrence_network import InterSystemRecurrenceNetwork
from .joint_recurrence_network import JointRecurrenceNetwork
from .joint_recurrence_plot import JointRecurrencePlot
from .recurrence_network import RecurrenceNetwork
from .recurrence_plot import RecurrencePlot
from .surrogates import Surrogates
from .visibility_graph import VisibilityGraph



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"
"""
Tests for the TsonisClimateNetwork class.
"""
import numpy as np

from pyunicorn.climate.climate_data import ClimateData
from pyunicorn.climate.tsonis import TsonisClimateNetwork


def test_str(capsys):
    print(TsonisClimateNetwork.SmallTestNetwork())
    out, err = capsys.readouterr()
    out_ref = "TsonisClimateNetwork:\n" + \
              "ClimateNetwork:\n" + \
              "GeoNetwork:\n" + \
              "SpatialNetwork:\n" + \
              "Network: undirected, 6 nodes, 6 links, link density 0.400." + \
              "\nGeographical boundaries:\n" + \
              "         time     lat     lon\n" + \
              "   min    0.0    0.00    2.50\n" + \
              "   max    9.0   25.00   15.00\n" + \
              "Threshold: 0.5\n" + \
              "Local connections filtered out: False\n" + \
              "Use only data points from winter months: False\n"
    assert out == out_ref


def test_SmallTestNetwork():
    res = TsonisClimateNetwork.SmallTestNetwork().adjacency
    exp = np.array([[0, 0, 1, 0, 1, 0], [0, 0, 0, 1, 0, 1],
                    [1, 0, 0, 0, 1, 0], [0, 1, 0, 0, 0, 1],
                    [1, 0, 1, 0, 0, 0], [0, 1, 0, 1, 0, 0]])
    assert np.allclose(res, exp, atol=1e-04)


def test_calculate_similarity_measure():
    res = TsonisClimateNetwork.SmallTestNetwork().calculate_similarity_measure(
        anomaly=ClimateData.SmallTestData().anomaly())
    exp = np.array([[1., -0.2538, -1., 0.2538, 1., -0.2538],
                    [-0.2538, 1., 0.2538, -1., -0.2538, 1.],
                    [-1., 0.2538, 1., -0.2538, -1., 0.2538],
                    [0.2538, -1., -0.2538, 1., 0.2538, -1.],
                    [1., -0.2538, -1., 0.2538, 1., -0.2538],
                    [-0.2538, 1., 0.2538, -1., -0.2538, 1.]],
                   dtype=np.float32)
    assert np.allclose(res, exp, atol=1e-04)


def test_correlation():
    res = TsonisClimateNetwork.SmallTestNetwork().correlation()
    exp = np.array([[1., 0.25377226, 1., 0.25377226, 1., 0.25377226],
                    [0.25377226, 1., 0.25377226, 1., 0.25377226, 1.],
                    [1., 0.25377226, 1., 0.25377226, 1., 0.25377226],
                    [0.25377226, 1., 0.25377226, 1., 0.25377226, 1.],
                    [1., 0.25377226, 1., 0.25377226, 1., 0.25377226],
                    [0.25377226, 1., 0.25377226, 1., 0.25377226, 1.]],
                   dtype=np.float32)
    assert np.allclose(res, exp, atol=1e-04)


def test_winter_only():
    res = TsonisClimateNetwork.SmallTestNetwork().winter_only()
    exp = False
    assert res == exp


def test_set_winter_only():
    net = TsonisClimateNetwork.SmallTestNetwork()
    net.set_winter_only(winter_only=False)

    res = net.n_links
    exp = 6
    assert res == exp


def test_correlation_weighted_average_path_length():
    res = TsonisClimateNetwork.SmallTestNetwork().\
            correlation_weighted_average_path_length()
    exp = 1.0
    assert np.isclose(res, exp, atol=1e-04)


def test_correlation_weighted_closeness():
    res = TsonisClimateNetwork.SmallTestNetwork().\
                correlation_weighted_closeness()
    exp = np.array([0.25, 0.25, 0.25, 0.25, 0.25, 0.25])
    assert np.allclose(res, exp, atol=1e-04)


def test_local_correlation_weighted_vulnerability():
    res = TsonisClimateNetwork.SmallTestNetwork().\
                local_correlation_weighted_vulnerability()
    exp = np.array([0., 0., 0., 0., 0., 0.])
    assert np.allclose(res, exp, atol=1e-04)



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

import numpy as np


BOOLTYPE = np.int8
INT8TYPE = np.int8
INT16TYPE = np.int16
INT32TYPE = np.int32
INT64TYPE = np.int64
FLOAT32TYPE = np.float32
FLOAT64TYPE = np.float64

ADJ = BOOLTYPE
MASK = BOOLTYPE
LAG = INT8TYPE
DEGREE = INT16TYPE
NODE = INT32TYPE
WEIGHT = FLOAT32TYPE
DWEIGHT = FLOAT64TYPE
FIELD = FLOAT32TYPE
DFIELD = FLOAT64TYPE


def to_cy(arr, ty):
    return arr.astype(dtype=ty, copy=True, order='c', casting='same_kind')



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"
"""
Simple tests for the InteractingNetworks class.
"""
import numpy as np

from pyunicorn.core.interacting_networks import InteractingNetworks


def test_RandomlyRewireCrossLinks():
    net = InteractingNetworks.SmallTestNetwork()
    rewired_net = net.RandomlyRewireCrossLinks(network=net,
                                               node_list1=[0, 3, 5],
                                               node_list2=[1, 2, 4], swaps=10.)

    res = rewired_net.degree()
    exp = [3, 3, 2, 2, 3, 1]
    assert (res == exp).all()

    res = rewired_net.cross_degree(node_list1=[0, 3, 5], node_list2=[1, 2, 4])
    exp = [1, 1, 0]
    assert (res == exp).all()


def test_internal_adjacency():
    net = InteractingNetworks.SmallTestNetwork()
    res = net.internal_adjacency([0, 3, 5])
    exp = np.array([[0, 1, 1], [1, 0, 0], [1, 0, 0]], dtype=np.int8)
    assert (res == exp).all()

    res = net.internal_adjacency([1, 2, 4])
    exp = np.array([[0, 1, 1], [1, 0, 1], [1, 1, 0]], dtype=np.int8)
    assert (res == exp).all()


def test_cross_adjacency():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.cross_adjacency([1, 2, 4], [0, 3, 5])
    exp = np.array([[0, 1, 0], [0, 0, 0], [1, 0, 0]])
    assert (res == exp).all()

    res = net.cross_adjacency([1, 2, 3, 4], [0, 5])
    exp = np.array([[0, 0], [0, 0], [1, 0], [1, 0]])
    assert (res == exp).all()


def test_cross_adjacency_sparse():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.cross_adjacency_sparse([1, 2, 4], [0, 3, 5])
    exp = [[0, 1, 0], [0, 0, 0], [1, 0, 0]]
    assert (res == exp).all()


def test_internal_link_attribute():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.internal_link_attribute("link_weights", [1, 2, 3])
    exp = np.array([[0., 2.3, 2.9],
                    [2.3, 0., 0.],
                    [2.9, 0., 0.]])
    assert np.allclose(res, exp, atol=1e-04)


def test_cross_link_attribute():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.cross_link_attribute("link_weights", [1, 2, 3], [0, 4])
    exp = np.array([[0., 2.7],
                    [0., 1.5],
                    [1.3, 0.]])
    assert np.allclose(res, exp, atol=1e-04)


def test_internal_path_lengths():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.internal_path_lengths([0, 3, 5], None)
    exp = np.array([[0., 1., 1.], [1., 0., 2.], [1., 2., 0.]])
    assert np.allclose(res, exp, atol=1e-04)

    res = net.internal_path_lengths([1, 2, 4], None)
    exp = np.array([[0., 1., 1.], [1., 0., 1.], [1., 1., 0.]])
    assert np.allclose(res, exp, atol=1e-04)


def test_cross_path_lengths():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.cross_path_lengths([0, 3, 5], [1, 2, 4], None)
    exp = np.array([[2., 2., 1.], [1., 2., 2.], [3., 3., 2.]])
    assert np.allclose(res, exp, atol=1e-04)

    res = net.cross_path_lengths([0, 5], [1, 2, 3, 4], None)
    exp = np.array([[2., 2., 1., 1.], [3., 3., 2., 2.]])
    assert np.allclose(res, exp, atol=1e-04)


def test_number_cross_links():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.number_cross_links([0, 3, 5], [1, 2, 4])
    exp = 2
    assert res == exp

    res = net.number_cross_links([0, 5], [1, 2, 3, 4])
    exp = 2
    assert res == exp


def test_total_cross_degree():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.total_cross_degree([0, 3, 5], [1, 2, 4])
    exp = 0.6667
    assert np.isclose(res, exp, atol=1e-04)

    res = net.total_cross_degree([0, 5], [1, 2, 3, 4])
    exp = 1.0
    assert np.isclose(res, exp, atol=1e-04)


def test_number_internal_links():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.number_internal_links([0, 3, 5])
    exp = 2
    assert res == exp

    res = net.number_internal_links([1, 2, 4])
    exp = 3
    assert res == exp


def test_cross_degree_density():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.cross_degree_density([0, 3, 5], [1, 2, 4])
    exp = np.array([0.33333333, 0.33333333, 0.])
    assert np.allclose(res, exp, atol=1e-04)


def test_cross_link_density():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.cross_link_density([0, 3, 5], [1, 2, 4])
    exp = 0.2222
    assert np.isclose(res, exp, atol=1e-04)

    res = net.cross_link_density([0, 5], [1, 2, 3, 4])
    exp = 0.25
    assert np.isclose(res, exp, atol=1e-04)


def test_internal_link_density():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.internal_link_density([0, 3, 5])
    exp = 0.6667
    assert np.isclose(res, exp, atol=1e-04)

    res = net.internal_link_density([1, 2, 3, 4])
    exp = 0.6667
    assert np.isclose(res, exp, atol=1e-04)


def test_internal_global_clustering():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.internal_global_clustering([0, 3, 5])
    exp = 0.0
    assert np.isclose(res, exp, atol=1e-04)

    res = net.internal_global_clustering([1, 2, 4])
    exp = 0.5556
    assert np.isclose(res, exp, atol=1e-04)


def test_cross_global_clustering():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.cross_global_clustering([0, 3, 5], [1, 2, 4])
    exp = 0.0
    assert np.isclose(res, exp, atol=1e-04)

    res = net.cross_global_clustering([2], [1, 3, 4])
    exp = 1.0
    assert np.isclose(res, exp, atol=1e-04)

    res = net.cross_global_clustering([3, 4], [1, 2])
    exp = 0.5
    assert np.isclose(res, exp, atol=1e-04)


def test_cross_global_clustering_sparse():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.cross_global_clustering_sparse([0, 3, 5], [1, 2, 4])
    exp = 0.0
    assert np.isclose(res, exp, atol=1e-04)

    res = net.cross_global_clustering_sparse([2], [1, 3, 4])
    exp = 1.0
    assert np.isclose(res, exp, atol=1e-04)

    res = net.cross_global_clustering_sparse([3, 4], [1, 2])
    exp = 0.5
    assert np.isclose(res, exp, atol=1e-04)


def test_cross_transitivity():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.cross_transitivity([0, 3, 5], [1, 2, 4])
    exp = 0.0
    assert np.isclose(res, exp, atol=1e-04)

    res = net.cross_transitivity([2], [1, 3, 4])
    exp = 1.0
    assert np.isclose(res, exp, atol=1e-04)

    res = net.cross_transitivity([3, 4], [1, 2])
    exp = 1.0


def test_cross_transitivity_sparse():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.cross_transitivity_sparse([0, 3, 5], [1, 2, 4])
    exp = 0.0
    assert np.isclose(res, exp, atol=1e-04)

    res = net.cross_transitivity_sparse([2], [1, 3, 4])
    exp = 1.0
    assert np.isclose(res, exp, atol=1e-04)

    res = net.cross_transitivity_sparse([3, 4], [1, 2])
    exp = 1.0
    assert np.isclose(res, exp, atol=1e-04)


def test_cross_average_path_length():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.cross_average_path_length([0, 3, 5], [1, 2, 4], None)
    exp = 2.0
    assert np.isclose(res, exp, atol=1e-04)

    res = net.cross_average_path_length([0, 5], [1, 2, 3, 4], None)
    exp = 2.0
    assert np.isclose(res, exp, atol=1e-04)


def test_internal_average_path_length():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.internal_average_path_length([0, 3, 5], None)
    exp = 1.3333
    assert np.isclose(res, exp, atol=1e-04)

    res = net.internal_average_path_length([1, 2, 4], None)
    exp = 1.0
    assert np.isclose(res, exp, atol=1e-04)


def test_average_cross_closeness():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.average_cross_closeness([0, 5], [1, 2, 3, 4])
    exp = 0.5333
    assert np.isclose(res, exp, atol=1e-04)


def test_global_efficiency():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.global_efficiency([0, 5], [1, 2, 3, 4])
    exp = 1.7143
    assert np.isclose(res, exp, atol=1e-04)


def test_cross_degree():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.cross_degree([0, 3, 5], [1, 2, 4])
    exp = np.array([1, 1, 0])
    assert (res == exp).all()

    res = net.cross_degree([1, 2, 4], [0, 3, 5])
    exp = np.array([1, 0, 1])
    assert (res == exp).all()

    res = net.cross_degree([1, 2, 3, 4], [0, 5])
    exp = np.array([0, 0, 1, 1])
    assert (res == exp).all()


def test_cross_indegree():
    net = InteractingNetworks.SmallDirectedTestNetwork()

    res = net.cross_indegree([1, 2], [0, 3, 4])
    exp = np.array([2, 1])
    assert (res == exp).all()


def test_cross_outdegree():
    net = InteractingNetworks.SmallDirectedTestNetwork()

    res = net.cross_outdegree([1, 2], [0, 3, 4])
    exp = np.array([1, 0])
    assert (res == exp).all()


def test_internal_degree():
    net = InteractingNetworks.SmallDirectedTestNetwork()

    res = net.internal_degree([0, 3, 5])
    exp = np.array([2, 1, 1])
    assert (res == exp).all()

    res = net.internal_degree([1, 2, 4])
    exp = np.array([2, 2, 2])
    assert (res == exp).all()


def test_internal_indegree():
    net = InteractingNetworks.SmallDirectedTestNetwork()

    res = net.internal_indegree([0, 1, 3])
    exp = np.array([0, 2, 1])
    assert (res == exp).all()


def test_internal_outdegree():
    net = InteractingNetworks.SmallDirectedTestNetwork()

    res = net.internal_outdegree([0, 1, 3])
    exp = np.array([2, 0, 1])
    assert (res == exp).all()


def test_cross_local_clustering():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.cross_local_clustering([0, 3, 5], [1, 2, 4])
    exp = np.array([0., 0., 0.])
    assert np.allclose(res, exp, atol=1e-04)

    res = net.cross_local_clustering([2], [1, 3, 4])
    exp = np.array([1.])
    assert np.allclose(res, exp, atol=1e-04)

    res = net.cross_local_clustering([3, 4], [1, 2])
    exp = np.array([0., 1.])
    assert np.allclose(res, exp, atol=1e-04)


def test_cross_local_clustering_sparse():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.cross_local_clustering_sparse([0, 3, 5], [1, 2, 4])
    exp = np.array([0., 0., 0.])
    assert np.allclose(res, exp, atol=1e-04)

    res = net.cross_local_clustering_sparse([2], [1, 3, 4])
    exp = np.array([1.])
    assert np.allclose(res, exp, atol=1e-04)

    res = net.cross_local_clustering_sparse([3, 4], [1, 2])
    exp = np.array([0., 1.])
    assert np.allclose(res, exp, atol=1e-04)


def test_cross_closeness():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.cross_closeness([0, 3, 5], [1, 2, 4], None)
    exp = np.array([0.6, 0.6, 0.375])
    assert np.allclose(res, exp, atol=1e-04)

    res = net.cross_closeness([0, 5], [1, 2, 3, 4], None)
    exp = np.array([0.66666667, 0.4])
    assert np.allclose(res, exp, atol=1e-04)


def test_internal_closeness():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.internal_closeness([0, 3, 5], None)
    exp = np.array([1., 0.66666667, 0.66666667])
    assert np.allclose(res, exp, atol=1e-04)

    res = net.internal_closeness([1, 2, 4], None)
    exp = np.array([1., 1., 1.])
    assert np.allclose(res, exp, atol=1e-04)


def test_cross_betweenness():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.cross_betweenness([2], [3, 5])
    exp = np.array([1., 1., 0., 0., 1., 0.])
    assert np.allclose(res, exp, atol=1e-04)

    res = net.cross_betweenness(range(0, 6), range(0, 6))
    exp = np.array([9., 3., 0., 2., 6., 0.])
    assert np.allclose(res, exp, atol=1e-04)


def test_internal_betweenness():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.internal_betweenness(range(0, 6))
    exp = np.array([9., 3., 0., 2., 6., 0.])
    assert np.allclose(res, exp, atol=1e-04)


def test_local_efficiency():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.local_efficiency([0, 5], [1, 2, 3, 4])
    exp = np.array([0.75, 0.41666667])
    assert np.allclose(res, exp, atol=1e-04)


def test_nsi_cross_degree():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.nsi_cross_degree([0, 1, 2], [3, 4, 5])
    exp = np.array([4.2, 2.6, 1.4])
    assert np.allclose(res, exp, atol=1e-04)

    res = net.nsi_cross_degree([0, 2, 5], [1, 4])
    exp = np.array([1.4, 2.2, 0.])
    assert np.allclose(res, exp, atol=1e-04)


def test_nsi_cross_mean_degree():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.nsi_cross_mean_degree([0, 1, 2], [3, 4, 5])
    exp = 2.5
    assert np.isclose(res, exp, atol=1e-04)

    res = net.nsi_cross_mean_degree([0, 2, 5], [1, 4])
    exp = 0.95
    assert np.isclose(res, exp, atol=1e-04)


def test_nsi_internal_degree():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.nsi_internal_degree([0, 3, 5])
    exp = np.array([3.4, 1.8, 2.2])
    assert np.allclose(res, exp, atol=1e-04)

    res = net.nsi_internal_degree([0, 1, 3, 5])
    exp = np.array([3.4, 2., 2.6, 2.2])
    assert np.allclose(res, exp, atol=1e-04)


def test_nsi_cross_local_clustering():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.nsi_cross_local_clustering([0, 1, 2], [3, 4, 5])
    exp = np.array([0.33786848, 0.50295858, 1.])
    assert np.allclose(res, exp, atol=1e-04)

    res = net.nsi_cross_local_clustering([0, 2, 5], [1, 4])
    exp = np.array([1., 1., 0.])
    assert np.allclose(res, exp, atol=1e-04)


def test_nsi_cross_closeness_centrality():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.nsi_cross_closeness_centrality([0, 1, 2], [3, 4, 5])
    exp = np.array([1., 0.56756757, 0.48837209])
    assert np.allclose(res, exp, atol=1e-04)

    res = net.nsi_cross_closeness_centrality([0, 2, 5], [1, 4])
    exp = np.array([0.73333333, 1., 0.42307692])
    assert np.allclose(res, exp, atol=1e-04)


def test_nsi_internal_closeness_centrality():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.nsi_internal_closeness_centrality([0, 3, 5])
    exp = np.array([1., 0.68, 0.73913043])
    assert np.allclose(res, exp, atol=1e-04)

    res = net.nsi_internal_closeness_centrality([0, 1, 3, 5])
    exp = np.array([0.84, 0.525, 0.72413793, 0.6])
    assert np.allclose(res, exp, atol=1e-04)


def test_nsi_cross_global_clustering():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.nsi_cross_global_clustering([0, 1, 2], [3, 4, 5])
    exp = 0.6688
    assert np.isclose(res, exp, atol=1e-04)


def test_nsi_internal_local_clustering():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.nsi_internal_local_clustering([1, 2, 3, 5])
    exp = np.array([0.73333333, 1., 1., 1.])
    assert np.allclose(res, exp, atol=1e-04)

    res = net.nsi_internal_local_clustering([0, 2, 4])
    exp = np.array([1., 1., 0.86666667])
    assert np.allclose(res, exp, atol=1e-04)


def test_nsi_cross_betweenness():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.nsi_cross_betweenness([0, 4, 5], [1, 3])
    exp = np.array([6.5333, 1.2, 0., 0.6769, 0.6769, 0.])
    assert np.allclose(res, exp, atol=1e-04)

    res = net.nsi_cross_betweenness([0, 1], [2, 3, 4, 5])
    exp = np.array([2.1333, 0., 0., 0.4923, 0.9209, 0.])
    assert np.allclose(res, exp, atol=1e-04)


def test_nsi_cross_edge_density():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.nsi_cross_edge_density([1, 2, 3], [0, 5])
    exp = 0.1091
    assert np.isclose(res, exp, atol=1e-04)

    res = net.nsi_cross_edge_density([0], [1, 4, 5])
    exp = 0.7895
    assert np.isclose(res, exp, atol=1e-04)


def test_nsi_cross_transitivity():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.nsi_cross_transitivity([1, 2], [0, 3, 4, 5])
    exp = 0.6352
    assert np.isclose(res, exp, atol=1e-04)

    res = net.nsi_cross_transitivity([0, 2, 3], [1])
    exp = 1.0
    assert np.isclose(res, exp, atol=1e-04)


def test_nsi_cross_average_path_length():
    net = InteractingNetworks.SmallTestNetwork()

    res = net.nsi_cross_average_path_length([0, 5], [1, 2, 4])
    exp = 3.3306
    assert np.isclose(res, exp, atol=1e-04)

    res = net.nsi_cross_average_path_length([1, 3, 4, 5], [2])
    exp = 0.376
    assert np.isclose(res, exp, atol=1e-04)



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
core
====

Provides classes for analyzing spatially embedded complex networks, handling
multivariate data and generating time series surrogates.

Related Publications
~~~~~~~~~~~~~~~~~~~~
[Donges2011a]_, [Heitzig2012]_, [Donges2012]_
"""

#
#  Import classes
#

from .network import Network, NetworkError, nz_coords
from .spatial_network import SpatialNetwork
from .geo_network import GeoNetwork
from .grid import Grid
from .geo_grid import GeoGrid
from .data import Data
from .interacting_networks import InteractingNetworks
from .netcdf_dictionary import NetCDFDictionary
from .resistive_network import ResNetwork

#
#  Set global constants
#

#  Mean earth radius in kilometers
EARTH_RADIUS = 6367.5
"""(float) - The earth's mean radius in kilometers."""
